# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests quantizer for pallas."""

# TODO(wppark): Remove this file. This is a temporary module before the
# official release of AQT quant / dequant API.

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax.v2 import aqt_tensor
from aqt.jax.v2.pallas import quantizer
import jax
import jax.numpy as jnp


QTensor = aqt_tensor.QTensor


class AqtPallasTest(parameterized.TestCase):

  @parameterized.parameters(
      (
          (1024, 1024),
          (1,),
          (1024, 1),
          jnp.float32
      ),
      (
          (1024, 1024),
          (1,),
          (1024, 1),
          jnp.float32
      ),
      (
          (1024, 1024),
          (0,),
          (1, 1024),
          jnp.float32
      ),
      (
          (10, 512, 1024),
          (1,),
          (10, 1, 1024),
          jnp.float32
      ),
      (
          (10, 512, 1024),
          (2,),
          (10, 512, 1),
          jnp.float32
      ),
      (
          (1024, 1024),
          (0,),
          (1, 1024),
          jnp.bfloat16
      ),
      (
          (10, 512, 1024),
          (1,),
          (10, 1, 1024),
          jnp.bfloat16
      ),
      (
          (10, 512, 1024),
          (2,),
          (10, 512, 1),
          jnp.bfloat16
      ),
  )
  def test_quant(
      self, tensor_shape, calibration_axes, expected_scale_shape, dtype
  ):
    key = jax.random.PRNGKey(0)
    x = jax.random.uniform(key, tensor_shape, minval=-3, maxval=3, dtype=dtype)
    qx = quantizer.quant(x, 8, calibration_axes)

    self.assertEqual(qx.qvalue.shape, x.shape)
    self.assertEqual(qx.qvalue.dtype, jnp.int8)
    self.assertEqual(qx.scale[0].shape, expected_scale_shape)
    # scale factor should be float32 regardless of the dtype of x.
    self.assertEqual(qx.scale[0].dtype, jnp.float32)


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Creates a sweep over all the leaderboard models.

Runs each model with three random seeds.
"""

import copy
from aqt.jax_legacy.jax.wmt_mlperf.hparams_configs.leaderboard import full_model_4bit_weights_and_auto_acts
from aqt.jax_legacy.jax.wmt_mlperf.hparams_configs.leaderboard import full_model_4bit_weights_only
from aqt.jax_legacy.jax.wmt_mlperf.hparams_configs.leaderboard import full_model_8bit_weights_and_auto_acts
from aqt.jax_legacy.jax.wmt_mlperf.hparams_configs.leaderboard import full_model_8bit_weights_only
from aqt.jax_legacy.jax.wmt_mlperf.hparams_configs.leaderboard import full_model_bfloat16
import ml_collections


def get_config():
  """Returns sweep configuration (see module docstring)."""
  sweep_config = ml_collections.ConfigDict()
  base_configs = [
      full_model_bfloat16.get_config(),
      full_model_8bit_weights_only.get_config(),
      full_model_8bit_weights_and_auto_acts.get_config(),
      full_model_4bit_weights_only.get_config(),
      full_model_4bit_weights_and_auto_acts.get_config()
  ]
  configs = []
  for base_config in base_configs:
    for seed in range(3):
      config = copy.deepcopy(base_config)
      config.random_seed = seed
      config.metadata.hyper_str = f"{config.metadata.hyper_str}_seed={seed}"
      configs.append(config)
  sweep_config.configs = configs
  return sweep_config



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions to load/save the hparams to/from a config dict."""

import dataclasses
import json
import os
import typing
from typing import Any, Dict, Optional, Type, TypeVar

from aqt.jax_legacy.jax import quant_config
from aqt.jax_legacy.jax import quantization
from aqt.jax_legacy.jax.flax import struct as flax_struct
from aqt.jax_legacy.jax.sparsity import sparsity_core_depr
import dacite
import jax
import ml_collections


T = TypeVar('T')

dataclass = (
    flax_struct.dataclass if not typing.TYPE_CHECKING else dataclasses.dataclass
)


@dataclass
class HParamsMetadata:
  """Metadata associated with an experiment configuration."""

  # Human-readable description of this hparams configuration. Mainly
  # useful for hand inspection of serialized JSON files.
  description: str

  # Creation time of the configuration in the format of seconds from epoch.
  # Used for versioning different hyperparameter settings for the same
  # model configuration.
  last_updated_time: Optional[float]

  # By default, it is used to name the model directory and label the
  # experiment in tensorboard.
  hyper_str: Optional[str] = None


# TODO(abdolrashidi): Add unit tests for the functions below.
def save_dataclass_to_disk(data, path: str):
  """Serializes the given dataclass to a JSON file on disk.

  Args:
    data: A dataclass instance.
    path: Path to save the dataclass to.
  """

  data_dict = dataclasses.asdict(data)
  with open(path, 'w') as file:
    json.dump(data_dict, file, indent=2)


def write_hparams_to_file_with_host_id_check(hparams,
                                             output_dir: Optional[str]):
  """Writes hparams to file for master host.

  Args:
    hparams: Hparams.
    output_dir: Output directory to save hparams to, saves as output_dir /
      'hparams_config.json.
  """
  if jax.host_id() == 0 and output_dir is not None:
    # The directory is usually created automatically by the time we reach here,
    # but on some training runs it appears not to be.
    # MakeDirs will create the directory if it doesn't already exist and is a
    # no-op if it already exists.


    os.makedirs(output_dir, exist_ok=True)
    save_dataclass_to_disk(hparams,
                           os.path.join(output_dir, 'hparams_config.json'))


def load_dataclass_from_dict(dataclass_name: Type[T],
                             data_dict: Dict[Any, Any]) -> T:
  """Converts parsed dictionary from JSON into a dataclass.

  Args:
    dataclass_name: Name of the dataclass.
    data_dict: Dictionary parsed from JSON.

  Returns:
    An instance of `dataclass` populated with the data from `data_dict`.
  """
  # Some fields in TrainingHParams are formal Python enums, but they are stored
  # as plain text in the json. Dacite needs to be given a list of which classes
  # to convert from a string into an enum. The classes of all enum values which
  # are stored in a TrainingHParams instance (directly or indirectly) should be
  # listed here. See https://github.com/konradhalas/dacite#casting.
  enum_classes = [
      quantization.QuantOps.ActHParams.InputDistribution,
      quantization.QuantType,
      quant_config.QuantGranularity,
      sparsity_core_depr.SparseType,
  ]
  data_dict = _convert_lists_to_tuples(data_dict)
  return dacite.from_dict(
      data_class=dataclass_name,
      data=data_dict,
      config=dacite.Config(cast=enum_classes))


T = TypeVar('T')


def _convert_lists_to_tuples(node: T) -> T:
  """Recursively converts all lists to tuples in a nested structure.

  Recurses into all lists and dictionary values referenced by 'node',
  converting all lists to tuples.

  Args:
    node: A Python structure corresponding to JSON (a dictionary, a list,
      scalars, and compositions thereof)

  Returns:
    A Python structure identical to the input, but with lists replaced by
      tuples.
  """

  if isinstance(node, dict):
    return {key: _convert_lists_to_tuples(value) for key, value in node.items()}
  elif isinstance(node, (list, tuple)):
    return tuple([_convert_lists_to_tuples(value) for value in node])
  else:
    return node


def load_dataclass_from_json(dataclass_name: Type[T], json_data: str) -> T:
  """Creates a dataclass instance from JSON.

  Args:
    dataclass_name: Name of the dataclass to deserialize the JSON into.
    json_data: A Python string containing JSON.

  Returns:
    An instance of 'dataclass' populated with the JSON data.
  """

  data_dict = json.loads(json_data)
  return load_dataclass_from_dict(dataclass_name, data_dict)




# TODO(shivaniagrawal): functionality `load_hparams_from_file` is created for a
# generic (model hparams independent) train_hparams class; either we should move
# towards shared TrainHparams or remove the following functionalities.
def load_hparams_from_config_dict(hparams_classname: Type[T],
                                  model_classname: Type[Any],
                                  config_dict: ml_collections.ConfigDict) -> T:
  """Loads hparams from a configdict, and populates its model object.

  Args:
    hparams_classname: Name of the hparams class.
    model_classname: Name of the model class within the hparams class
    config_dict: A config dict mirroring the structure of hparams.

  Returns:
    An instance of 'hparams_classname' populated with the data from
    'config_dict'.
  """

  hparams = load_dataclass_from_config_dict(hparams_classname, config_dict)
  hparams.model_hparams = load_dataclass_from_dict(model_classname,
                                                   hparams.model_hparams)
  return hparams


def load_dataclass_from_config_dict(
    dataclass_name: Type[T], config_dict: ml_collections.ConfigDict) -> T:
  """Creates a dataclass instance from a configdict.

  Args:
    dataclass_name: Name of the dataclass to deserialize the configdict into.
    config_dict: A config dict mirroring the structure of 'dataclass_name'.

  Returns:
    An instance of 'dataclass_name' populated with the data from 'config_dict'.
  """

  # We convert the config dicts to JSON instead of a dictionary to force all
  # recursive field references to fully resolve in a way that Dacite can
  # consume.
  json_data = config_dict.to_json()
  return load_dataclass_from_json(dataclass_name, json_data)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Accurate Quantized Training ops.

Refer to //knowledge/cerebra/catalyst/tensorflow/aqt/aqt_ops.py for a core of
the quantization algorithm in AQTp.
"""
# TODO(jihwanlee): Create a shared doc, referred to by both TF and Jax.


from aqt.jax import aqt_conv_general
from aqt.jax import aqt_dot_general
from aqt.jax import aqt_matmul


aqt_conv_general_dilated = aqt_conv_general.conv_general_dilated
aqt_dot = aqt_dot_general.dot
aqt_dot_general = aqt_dot_general.dot_general
aqt_matmul = aqt_matmul.matmul



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Flax implementation of PokeBNN."""

# pylint: disable=unnecessary-lambda
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=using-constant-test

from typing import Any
# from aqt.jax import flax_layers as aqt_flax_layers
# from aqt.jax import quant_config
from aqt.jax_legacy.jax import flax_layers as aqt_flax_layers
from aqt.jax_legacy.jax import quant_config
from flax import linen as nn
import jax.numpy as jnp


class DPReLU(nn.Module):
  """Implements: dprelu(x) = slope * (x - bias_x) - bias_y.

     where slope depends on the sign of (x - bias_x).
  """

  @nn.compact
  def __call__(self, x):
    params_shape = (x.shape[-1],)

    bias_init_fn = lambda key, shape: jnp.ones(shape) * 0.0
    bias_x = self.param('bias_x', bias_init_fn, params_shape)
    bias_y = self.param('bias_y', bias_init_fn, params_shape)

    neg_slope_init_fn = lambda key, shape: jnp.ones(shape) * 0.25
    pos_slope_init_fn = lambda key, shape: jnp.ones(shape)
    neg_slope = self.param('neg_slope', neg_slope_init_fn, params_shape)
    pos_slope = self.param('pos_slope', pos_slope_init_fn, params_shape)

    x = x - bias_x
    slope = jnp.where(x >= 0, pos_slope, neg_slope)
    x = slope * x - bias_y
    return jnp.asarray(x, jnp.bfloat16)


class PokeBNN(nn.Module):
  # All the configurations goes through hparams dictionary.
  num_classes: int
  hparams: Any
  # Dynamic information on whether we should be quantizing already.
  dynamic_context: quant_config.DynamicContext
  train: bool
  # Jax way of multi-machine distribution
  paxis_name: Any
  # Logs all the ops with computational contents. Used for ACE calculation.
  op_log = []

  # Instruments other ops to log shaps and their arguments to self.op_log.
  def instr(self, name, op, other):
    def f(x, **kwargs):
      y = op(x, **kwargs)
      self.op_log.append((name, x.shape, y.shape, other))
      return y
    return f

  # All the common setting, plumbing and instrumentation.
  def batch_norm(self, **kwargs):
    op = nn.BatchNorm(
        use_running_average=not self.train,
        momentum=0.9,
        epsilon=1e-5,
        dtype=jnp.bfloat16,
        **kwargs)
    return self.instr('batch_norm', op, kwargs)

  # All the common setting, plumbing and instrumentation.
  def conv(self, use_bias=False, **kwargs):
    # ConvAqt implements quantization calibration for 8-, 4- and 1-bit convs.
    op = aqt_flax_layers.ConvAqt(
        use_bias=use_bias,
        dtype=jnp.bfloat16,
        dynamic_context=self.dynamic_context,
        paxis_name=self.paxis_name,
        train=self.train,
        **kwargs)
    return self.instr('conv', op, kwargs)

  # All the common setting, plumbing and instrumentation.
  def dense(self, **kwargs):
    op = aqt_flax_layers.DenseAqt(
        dtype=jnp.bfloat16,
        train=self.train,
        dynamic_context=self.dynamic_context,
        paxis_name=self.paxis_name,
        hparams=self.hparams.dense_layer,
        **kwargs)
    return self.instr('dense', op, (kwargs, self.hparams.dense_layer))

  # Instrumentation of DPReLU.
  def dprelu(self):
    op = DPReLU()
    return self.instr('dprelu', op, ())

  # reshape_add as described in the paper + some manual detailed instrumentation
  def reshape_add(self, r, yy, method, name):
    """Match the dimension for residuals and add to activations."""
    if r is None:
      return yy
    self.op_log.append(('reshape_add', r.shape, name, ()))
    out_features = yy.shape[3]
    in_features = r.shape[-1]
    if in_features > out_features:
      num_ch_avg = in_features // out_features
      r = r[:, :, :, 0:out_features * num_ch_avg]
      dim_nwh = r.shape[0:3]
      r = jnp.reshape(r, dim_nwh + (out_features, num_ch_avg))
      self.op_log.append(
          ('reshape_add:average axis=4; before', r.shape, (), ()))
      r = jnp.average(r, axis=4)
    elif in_features < out_features:
      ch_mult = out_features // in_features
      assert method in ['tile', 'zeropad']
      if method == 'tile':
        self.op_log.append(
            ('reshape_add:tile axis=3; before', r.shape, ch_mult, ()))
        r = jnp.tile(r, reps=(1, 1, 1, ch_mult))
      if method == 'zeropad':
        pad_size = out_features - in_features
        pad_width = ((0, 0), (0, 0), (0, 0), (0, pad_size))
        self.op_log.append(
            ('reshape_add:zeropad axis=3; before', r.shape, pad_size, ()))
        r = jnp.pad(r, pad_width=pad_width, mode='constant')
    if yy.shape != r.shape:
      r = nn.avg_pool(r, window_shape=(3, 3), padding='SAME', strides=(2, 2))
      self.op_log.append(
          ('reshape_add:avg_pool3x3, st=2x2; after', r.shape, (), ()))
    self.op_log.append(('reshape_add:add ' + method, r.shape, (), ()))
    return yy + r

  def poke_init(self, x):
    # No input pixel is used twice.
    x = self.conv(
        features=self.hparams.init_group,
        kernel_size=(4, 4),
        strides=(4, 4),
        padding='VALID',
        name='init_conv',
        hparams=self.hparams.conv_init)(
            x)
    x = self.batch_norm(name='init_bn')(x)
    x = self.dprelu()(x)
    x = self.conv(
        features=64,
        kernel_size=(3, 3),
        strides=(1, 1),
        padding='SAME',
        feature_group_count=self.hparams.init_group,
        name='init_conv_group',
        hparams=self.hparams.conv_init)(
            x)
    x = self.batch_norm(name='init_bn_group')(x)
    x = self.dprelu()(x)
    return x

  def se(self, x, features, name):
    proj_len = x.shape[-1] // 8
    hparams = self.hparams.residual_blocks[0].conv_se
    self.op_log.append(
        ('se:global_avg_pool axis = 1,2; before', x.shape, (), ()))
    # TODO(yichi): Test that this line replaces the next two and do it.
    # s = jnp.mean(x, axis=(1, 2), keepdims=True)
    s = jnp.apply_over_axes(jnp.mean, x, (1, 2))
    s = s.reshape(-1, 1, 1, x.shape[-1])

    hparams.quant_act.input_distribution = 'symmetric'  # use int4
    s = self.conv(
        use_bias=True,
        kernel_size=(1, 1),
        features=proj_len,
        hparams=hparams,
        name='sefc1' + name)(
            s)
    s = nn.relu(s)
    self.op_log.append(('se:relu', s.shape, (), ()))

    hparams.quant_act.input_distribution = 'positive'  # use uint4, save 1 bit
    s = self.conv(
        use_bias=True,
        kernel_size=(1, 1),
        features=features,
        hparams=hparams,
        name='sefc2' + name)(
            s)
    s = jnp.minimum(jnp.maximum(s + 3, 0), 6.) / 6
    self.op_log.append(('se:relu6 shifted', s.shape, (), ()))

    return s

  def poke_conv(self, x, r1, name, features, kernel_size, strides):
    r = x
    x = self.conv(
        features=features,
        kernel_size=kernel_size,
        name='conv-' + name,
        strides=strides,
        hparams=self.hparams.residual_blocks[0].conv_1)(
            x)
    # ResNet does this. It is important.
    scale_init = nn.initializers.zeros if name[
        -1] == '3' else nn.initializers.ones
    x = self.batch_norm(name='bn' + name, scale_init=scale_init)(x)
    x = self.reshape_add(r, x, 'zeropad', 'local-shortcut-' + name)
    x = self.reshape_add(r1, x, 'tile', 'long-shortcut-' + name)
    x = self.dprelu()(x)
    s = self.se(x=r, features=x.shape[3], name='-' + name)
    x = x * s
    self.op_log.append(('se:poke_relu scale', x.shape, s.shape, ()))

    x = self.batch_norm(name='bn_split' + name)(x)
    return x

  def bottleneck_block(self, x, name, hparams, features, strides):
    r1 = x
    x = self.poke_conv(
        x,
        r1=None,
        name=name + '-1',
        features=features,
        kernel_size=(1, 1),
        strides=(1, 1),
    )
    x = self.poke_conv(
        x,
        r1=None,
        name=name + '-2',
        features=features,
        kernel_size=(3, 3),
        strides=strides,
    )
    x = self.poke_conv(
        x,
        r1=r1,
        name=name + '-3',
        features=features * 4,
        kernel_size=(1, 1),
        strides=(1, 1),
    )
    return x

  @nn.compact
  def __call__(self, x):
    self.op_log[:] = []
    # self.ace[0] = 0

    x = self.poke_init(x)

    # This list was generated from ResNet-50, and probably can be improved.
    for i, strides, features in [
        (0, (1, 1), 64),
        (1, (1, 1), 64),
        (2, (1, 1), 64),
        (3, (2, 2), 128),
        (4, (1, 1), 128),
        (5, (1, 1), 128),
        (6, (1, 1), 128),
        (7, (2, 2), 256),
        (8, (1, 1), 256),
        (9, (1, 1), 256),
        (10, (1, 1), 256),
        (11, (1, 1), 256),
        (12, (1, 1), 256),
        (13, (2, 2), 512),
        (14, (1, 1), 512),
        (15, (1, 1), 512),
    ]:
      x = self.bottleneck_block(
          x,
          name=f'resblock{i}',
          features=int(features * self.hparams.filter_multiplier),
          hparams=self.hparams.residual_blocks[i],
          strides=strides,
      )

    self.op_log.append(
        ('pokebnn: global mean axis=1,2; before', x.shape, (), ()))
    x = jnp.mean(x, axis=(1, 2))

    x = self.dense(features=self.num_classes)(x, padding_mask=None)
    x = jnp.asarray(x, jnp.bfloat16)
    return x

  # This function calculates all the stats for the 'elementwise' appendix.
  def size_stats(self):
    conv_ace = 0
    dense_ace = 0
    batch_norm_sops = 0  # sops means scalar ops (per pixel, per feature)
    bprelu_sops = 0
    reshape_add_ch_shrink_sops = 0
    reshape_add_ch_shrink_mul_sops = 0
    reshape_add_avg_pool_sops = 0
    reshape_add_block_residual_sops = 0
    reshape_add_local_residual_sops = 0
    se_global_pool_add_sops = 0
    se_global_pool_mul_sops = 0
    se_relu_sops = 0
    se_relu6_sops = 0
    se_mul_sops = 0
    pokebnn_global_pool_sops = 0

    dense_stats = {}
    conv_stats = {}
    residual_params = {}

    for op in self.op_log:
      name = op[0]
      if False:
        pass

      elif name == 'conv':
        _, xs, ys, kwargs = op
        out_pixels = ys[1] * ys[2]
        ks0 = kwargs['kernel_size'][0]
        ks1 = kwargs['kernel_size'][1]
        weight_prec = kwargs['hparams'].weight_prec
        act_prec = kwargs['hparams'].quant_act.prec
        # fgc takes into account depth-wise layer
        fgc = kwargs.get('feature_group_count', 1)
        dot_length = ks0 * ks1 * (xs[3] / fgc)
        adders = weight_prec * act_prec
        this_conv_ace = out_pixels * dot_length * ys[3] * adders
        conv_ace += this_conv_ace

        conv_name = kwargs['name']
        conv_stats[conv_name] = {
            'input_shape': xs[1:],  # H, W, C
            'input_prec': act_prec,
            'weight_shape': (ks0, ks1, xs[3] // fgc, ys[3]),  # ks, ks, Ci, Co
            'weight_prec': weight_prec,
            'conv_ace': this_conv_ace,
        }

      elif name == 'dense':
        _, xs, ys, (kwargs, hparams) = op  # pytype: disable=bad-unpacking
        adders = hparams.weight_prec * hparams.quant_act.prec
        dense_ace += xs[1] * ys[1] * adders
        dense_stats[name] = {
            'input_shape': xs[1],
            'input_prec': hparams.quant_act.prec,
            'weight_shape': (xs[1], ys[1]),
            'weight_prec': hparams.weight_prec,
            'output_shape': ys[1],
            'dense_ace': dense_ace,
        }

      elif name == 'dprelu':
        _, xs, ys, _ = op
        assert xs == ys
        bprelu_sops += xs[1] * xs[2] * xs[3]

      # During the training BN has two bf16 multiplications and two additions.
      # In principle it can be folded to one of each.
      # The energy would be: (210 + 110) fJ or 256 ACE per each sop.
      elif name == 'batch_norm':
        _, xs, ys, _ = op
        assert xs == ys
        batch_norm_sops += xs[1] * xs[2] * xs[3]

      elif name == 'reshape_add:average axis=4; before':
        _, sh, _, _ = op
        # height, width, #groups, group size
        reshape_add_ch_shrink_sops += sh[1] * sh[2] * sh[3] * sh[4]
        reshape_add_ch_shrink_mul_sops += sh[1] * sh[2] * sh[3]

      elif name == 'reshape_add:tile axis=3; before':
        # _, sh, ch_multiplier, _ = op
        pass

      elif name == 'reshape_add:zeropad axis=3; before':
        # _, sh, pad_size, _ = op
        pass

      elif name == 'reshape_add:avg_pool3x3, st=2x2; after':
        _, sh, _, _ = op
        reshape_add_avg_pool_sops += 3 * 3 * sh[1] * sh[2] * sh[3]

      elif name == 'reshape_add:add tile':
        _, sh, _, _ = op
        reshape_add_block_residual_sops += sh[1] * sh[2] * sh[3]

      # we don't need to add zeros during the computation on hw
      elif name == 'reshape_add:add zeropad':
        _, sh, _, _ = op
        reshape_add_local_residual_sops += sh[1] * sh[2] * sh[3]

      elif name == 'se:global_avg_pool axis = 1,2; before':
        _, sh, _, _ = op
        se_global_pool_add_sops += sh[1] * sh[2] * sh[3]
        se_global_pool_mul_sops += sh[3]

      elif name == 'se:relu':
        _, sh, _, _ = op
        se_relu_sops += sh[1] * sh[2] * sh[3]

      elif name == 'se:relu6 shifted':
        _, sh, _, _ = op
        se_relu6_sops += sh[1] * sh[2] * sh[3]

      elif name == 'se:poke_relu scale':
        _, x_sh, _, _ = op
        # We can use s_sh (SE pre-pixel-broadcast) because in principle it can
        # be fused with the following BN.
        se_mul_sops += x_sh[1] * x_sh[2] * x_sh[3]

      elif name == 'pokebnn: global mean axis=1,2; before':
        _, sh, _, _ = op
        pokebnn_global_pool_sops += sh[1] * sh[2] * sh[3]

      elif name == 'reshape_add':
        _, r_shape, shortcut_name, _ = op
        residual_params[shortcut_name] = {
            'shape': r_shape[1:],
            'params': r_shape[1] * r_shape[2] * r_shape[3],
        }

      else:
        assert False

    return {
        'conv_ace': conv_ace,
        'dense_ace': dense_ace,
        'batch_norm_sops': batch_norm_sops,
        'bprelu_sops': bprelu_sops,
        'reshape_add_ch_shrink_sops': reshape_add_ch_shrink_sops,
        'reshape_add_ch_shrink_mul_sops': reshape_add_ch_shrink_mul_sops,
        'reshape_add_avg_pool_sops': reshape_add_avg_pool_sops,
        'reshape_add_block_residual_sops': reshape_add_block_residual_sops,
        'reshape_add_local_residual_sops': reshape_add_local_residual_sops,
        'se_global_pool_add_sops': se_global_pool_add_sops,
        'se_global_pool_mul_sops': se_global_pool_mul_sops,
        'se_relu_sops': se_relu_sops,
        'se_relu6_sops': se_relu6_sops,
        'se_mul_sops': se_mul_sops,
        'pokebnn_global_pool_sops': pokebnn_global_pool_sops,
        'conv_stats': conv_stats,
        'residual_params': residual_params,
        'dense_stats': dense_stats,
    }


def create_pokebnn(hparams, train, **kwargs):
  return PokeBNN(
      num_classes=1000,
      hparams=hparams,
      dynamic_context=quant_config.DynamicContext(
          update_bounds=False, quantize_weights=True),
      train=train,
      paxis_name='batch',
      **kwargs)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Approximate BLEU score.

MLPerf version of BLEU calculation.
Tries to match SacreBLEU metric reasonably well.

Refs:
    tokenizer at:
    https://github.com/tensorflow/models/blob/master/official/transformer/utils/tokenizer.py
    original preprocessing tokenizer:
    https://github.com/moses-smt/mosesdecoder/blob/master/scripts/generic/mteval-v14.pl#L954-L983
    original t2t code:
    https://github.com/tensorflow/tensor2tensor/blob/master/tensor2tensor/utils/bleu_hook.py

Usage:
    refs = '''food bar brown cow
    blee bloo dog sat
    or please take me out
    '''
    hyps = '''foo bar brown cow
    blee bloo dog sit
    please do take me out
    '''
    bleu_local(refs.split("\n"), hyps.split("\n"))  # 39.65
"""

import collections
import math
import re
import sys
import unicodedata
import numpy as np
import six
import tensorflow.compat.v2 as tf


class UnicodeRegex(object):
  """Ad-hoc hack to recognize all punctuation and symbols."""

  def __init__(self):
    punctuation = self.property_chars("P")
    self.nondigit_punct_re = re.compile(r"([^\d])([" + punctuation + r"])")
    self.punct_nondigit_re = re.compile(r"([" + punctuation + r"])([^\d])")
    self.symbol_re = re.compile("([" + self.property_chars("S") + "])")

  def property_chars(self, prefix):
    return "".join(
        six.unichr(x)
        for x in range(sys.maxunicode)
        if unicodedata.category(six.unichr(x)).startswith(prefix))


uregex = UnicodeRegex()


def bleu_tokenize(string):
  r"""Tokenize a string following the official BLEU implementation.

  See https://github.com/moses-smt/mosesdecoder/'
           'blob/master/scripts/generic/mteval-v14.pl#L954-L983
  In our case, the input string is expected to be just one line
  and no HTML entities de-escaping is needed.
  So we just tokenize on punctuation and symbols,
  except when a punctuation is preceded and followed by a digit
  (e.g. a comma/dot as a thousand/decimal separator).

  Note that a number (e.g. a year) followed by a dot at the end of sentence
  is NOT tokenized, i.e. the dot stays with the number because
  `s/(\p{P})(\P{N})/ $1 $2/g` does not match this case (unless we add a
  space after each sentence). However, this error is already in the
  original mteval-v14.pl and we want to be consistent with it.

  Args:
    string: the input string

  Returns:
    a list of tokens
  """
  string = uregex.nondigit_punct_re.sub(r"\1 \2 ", string)
  string = uregex.punct_nondigit_re.sub(r" \1 \2", string)
  string = uregex.symbol_re.sub(r" \1 ", string)
  return string.split()


def _get_ngrams(segment, max_order):
  """Extracts all n-grams up to a given maximum order from an input segment.

  Args:
    segment: text segment from which n-grams will be extracted.
    max_order: maximum length in tokens of the n-grams returned by this methods.

  Returns:
    The Counter containing all n-grams up to max_order in segment
    with a count of how many times each n-gram occurred.
  """
  ngram_counts = collections.Counter()
  for order in range(1, max_order + 1):
    for i in range(0, len(segment) - order + 1):
      ngram = tuple(segment[i:i + order])
      ngram_counts[ngram] += 1
  return ngram_counts


def compute_bleu(reference_corpus,
                 translation_corpus,
                 max_order=4,
                 use_bp=True):
  """Computes BLEU score of translated segments against one or more references.

  Args:
    reference_corpus: list of references for each translation. Each reference
      should be tokenized into a list of tokens.
    translation_corpus: list of translations to score. Each translation should
      be tokenized into a list of tokens.
    max_order: Maximum n-gram order to use when computing BLEU score.
    use_bp: boolean, whether to apply brevity penalty.

  Returns:
    BLEU score.
  """
  reference_length = 0
  translation_length = 0
  bp = 1.0
  geo_mean = 0

  matches_by_order = [0] * max_order
  possible_matches_by_order = [0] * max_order
  precisions = []

  for (references, translations) in zip(reference_corpus, translation_corpus):
    reference_length += len(references)
    translation_length += len(translations)
    ref_ngram_counts = _get_ngrams(references, max_order)
    translation_ngram_counts = _get_ngrams(translations, max_order)

    overlap = dict((ngram, min(count, translation_ngram_counts[ngram]))
                   for ngram, count in ref_ngram_counts.items())

    for ngram in overlap:
      matches_by_order[len(ngram) - 1] += overlap[ngram]
    for ngram in translation_ngram_counts:
      possible_matches_by_order[len(ngram) -
                                1] += translation_ngram_counts[ngram]
  precisions = [0] * max_order
  smooth = 1.0
  for i in range(0, max_order):
    if possible_matches_by_order[i] > 0:
      precisions[i] = matches_by_order[i] / possible_matches_by_order[i]
      if matches_by_order[i] > 0:
        precisions[i] = matches_by_order[i] / possible_matches_by_order[i]
      else:
        smooth *= 2
        precisions[i] = 1.0 / (smooth * possible_matches_by_order[i])
    else:
      precisions[i] = 0.0

  if max(precisions) > 0:
    p_log_sum = sum(math.log(p) for p in precisions if p)
    geo_mean = math.exp(p_log_sum / max_order)

  if use_bp:
    if not reference_length:
      bp = 1.0
    else:
      ratio = translation_length / reference_length
      if ratio <= 0.0:
        bp = 0.0
      elif ratio >= 1.0:
        bp = 1.0
      else:
        bp = math.exp(1 - 1. / ratio)
  bleu = geo_mean * bp
  return np.float32(bleu)


def bleu_local(ref_lines, hyp_lines, case_sensitive=False):
  """Compute BLEU for two lists of reference and hypothesis translations."""
  if len(ref_lines) != len(hyp_lines):
    raise ValueError("Reference and translation lists have different "
                     "numbers of lines.")
  if not case_sensitive:
    ref_lines = [x.lower() for x in ref_lines]
    hyp_lines = [x.lower() for x in hyp_lines]
  ref_tokens = [bleu_tokenize(x) for x in ref_lines]
  hyp_tokens = [bleu_tokenize(x) for x in hyp_lines]
  return compute_bleu(ref_tokens, hyp_tokens) * 100


def bleu_wrapper(ref_filename, hyp_filename, case_sensitive=False):
  """Compute BLEU for two files (reference and hypothesis translation)."""
  ref_lines = tf.io.gfile.GFile(ref_filename).read().strip().splitlines()
  hyp_lines = tf.io.gfile.GFile(hyp_filename).read().strip().splitlines()
  return bleu_local(ref_lines, hyp_lines, case_sensitive=case_sensitive)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Quantized matmul."""

import builtins

import jax
from jax import lax
import jax.numpy as jnp
import numpy as np


shape = np.shape
ndim = np.ndim
_max = builtins.max


def matmul(a: jnp.ndarray,  #
           b: jnp.ndarray,
           *,
           precision=None,
           dot_general=lax.dot_general) -> jnp.ndarray:
  """Quantized jax.numpy.matmul.

  Args:
    a: Left-hand side of the matmul.
    b: Right-hand side of the matmul.
    precision: Indicates precision of a and b.
    dot_general: lax.dot_general by default. To use quantized matmul, the
      wrapper of aqt_dot_general in which TQs and `train` flag are provided
      should be passed into this function.

  Returns:
    An array containing the result with the same dtype as 'a' and 'b'.
  """
  arraylike = (jax.Array, np.ndarray)
  if not isinstance(a, arraylike) or not isinstance(b, arraylike):
    raise TypeError(f"matmul requires array-like arguments, got {a} and {b}")
  for i, x in enumerate((a, b)):
    if ndim(x) < 1:
      msg = (f"matmul input operand {i} must have ndim at least 1, "
             f"but it has ndim {ndim(x)}")
      raise ValueError(msg)

  dtype = jnp.result_type(a.dtype, b.dtype)
  a = a.astype(dtype)
  b = b.astype(dtype)

  a_is_mat, b_is_mat = (ndim(a) > 1), (ndim(b) > 1)
  a_batch_dims = shape(a)[:-2] if a_is_mat else ()
  b_batch_dims = shape(b)[:-2] if b_is_mat else ()
  num_batch_dims = _max(len(a_batch_dims), len(b_batch_dims))
  a_batch_dims = (None,) * (num_batch_dims - len(a_batch_dims)) + a_batch_dims
  b_batch_dims = (None,) * (num_batch_dims - len(b_batch_dims)) + b_batch_dims

  # Dimensions to squeeze from the inputs.
  a_squeeze = []
  b_squeeze = []

  # Positions of batch dimensions in squeezed inputs.
  a_batch = []
  b_batch = []

  # Desired index in final output of each kind of dimension, in the order that
  # aqt_dot_general will emit them.
  idx_batch = []
  idx_a_other = []  # other = non-batch, non-contracting.
  idx_b_other = []
  for i, (ba, bb) in enumerate(zip(a_batch_dims, b_batch_dims)):
    if ba is None:
      idx_b_other.append(i)
    elif bb is None:
      idx_a_other.append(i)
    elif ba == 1:
      idx_b_other.append(i)
      a_squeeze.append(len(idx_batch) + len(idx_a_other) + len(a_squeeze))
    elif bb == 1:
      idx_a_other.append(i)
      b_squeeze.append(len(idx_batch) + len(idx_b_other) + len(b_squeeze))
    elif ba == bb:
      a_batch.append(len(idx_batch) + len(idx_a_other))
      b_batch.append(len(idx_batch) + len(idx_b_other))
      idx_batch.append(i)
    else:
      raise ValueError("Incompatible shapes for matmul arguments: {} and {}"
                       .format(shape(a), shape(b)))

  if a_is_mat: idx_a_other.append(num_batch_dims)
  if b_is_mat: idx_b_other.append(num_batch_dims + a_is_mat)
  perm = np.argsort(np.concatenate([idx_batch, idx_a_other, idx_b_other]))

  a = lax.squeeze(a, tuple(a_squeeze))
  b = lax.squeeze(b, tuple(b_squeeze))
  out = dot_general(
      a,
      b, (((ndim(a) - 1,), (ndim(b) - 1 - b_is_mat,)), (a_batch, b_batch)),
      precision=precision)
  return lax.transpose(out, perm)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Transpose functions."""

# Lingo in this file:
#
# - lhs(rhs) - left(right) hand side of a binary operation
# - ca - contraction axes
# - ba - batch axes
# - ra - remaining axes

from typing import Sequence
from aqt.jax.v2 import utils
import jax
import jax.numpy as jnp


def transpose(t: jax.Array, axes: list[utils.AxisIdx]):
  """Transpose tensor through reshape if possible.

  This is a workaround to use AQT inside pallas kernel. Pallas support
  transposing only 2D tensor. However, tensor with a specific pattern can be
  transposed through reshaping. This pattern is observed a lot in scale factor
  of QTensor.

  If axes with dimensions larger than one are ordered in ascending order, then
  the tensor can be transposed by reshaping. For example,

  Let's say the shape of tensor `x` is [10, 1, 1]. Then, there are only one
  axis with dimension larger than one. The tensor can be transposed by
  reshaping no matter what the order of permutation, e.g., `jnp.transpose(x, [1,
  2, 0]) == jnp.reshape(x, [x.shape[i] for i in [1, 2, 0]])`.

  Another example is `x` with shape [10, 1, 30]. Then, there are two axes with
  dimension larger than one. The tensor can be transposed by reshaping if
  tensor is permuted while maintaining the order of axes with dimension larger
  than one, i.e., the first axes remains in front of third axis after
  permutation, e.g.,  `jnp.transpose(x, [0, 2, 1]) == jnp.reshape(x, [x.shape[i]
  for i in [0, 2, 1]])`.

  Args:
    t: input tensor.
    axes: transpose axes.

  Returns:
    A transposed tensor.

  Example:
    >>> x = np.random.randn([10, 1, 1]) # x is transposable by reshaping.
    >>> axes = [2, 1, 0]
    >>> y = jnp.transpose(x, axes)
    >>> (jnp.reshape(x, y.shape) == y).all()
    True
    >>> (transpose(x, axes) == y).all()

    >>> x = np.random.randn([10, 20, 30]) # x is not transposable by reshaping.
    >>> axes = [2, 1, 0]
    >>> y = jnp.transpose(x, axes)
    >>> (jnp.reshape(x, y.shape) == y).all()
    False
    >>> (transpose(x, axes) == y).all()
    True
  """

  # To check if a tensor's shape is not polymorphic, validate that the shape
  # consists of only integers.
  if all(map(lambda i: isinstance(i, int), t.shape)):
    axes_bigger_than_one = [i for i, size in enumerate(t.shape) if size > 1]
    permutation_axes_bigger_than_one = [
        axes.index(i) for i in axes_bigger_than_one
    ]
    # Check if axes bigger than one are ordered in ascending order.
    reshapable = (
        sorted(permutation_axes_bigger_than_one)
        == permutation_axes_bigger_than_one
    )
  else:
    reshapable = False

  if reshapable:
    t = jnp.reshape(t, (t.shape[i] for i in axes))
  else:
    t = jax.lax.transpose(t, axes)
  return t


def _scale_trans(x, ca, ba):
  """Transposes x to output dimension order."""
  ca = list(ca)
  ba = list(ba)
  for i in ca:
    assert x.shape[i] == 1, (x.shape, ca, ba)
  ra = utils.get_remaining_axes(x.ndim, ca, ba)
  x = transpose(x, ba + ra + ca)
  # TODO(lew): x = jnp.squeeze(x, axis=range(len(ba+ra): len(x.shape))
  shape_ba = x.shape[: len(ba)]
  shape_ra = x.shape[len(ba) : len(x.shape) - len(ca)]
  # Will need to add additional axes (size 1) for the other shape_ra
  x = x.reshape(shape_ba + shape_ra)
  return x


def lhs_scale_transpose_to_output(
    lhs_scale, dimension_numbers, lhs_shape, rhs_shape
):
  """Transposes lhs_scale to output dimension order."""
  if lhs_scale is None:
    return None
  # The axis order in out is as follows: batch, lhs_ra, rhs_ra
  # - batch axes order is uniquely determined by either lhs_ba or rhs_ba
  # - contraction axes ca disappear from the output
  # - order of the remaining axes (ra) is preserved.
  (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = dimension_numbers
  qlhs_scale_t = _scale_trans(lhs_scale, lhs_ca, lhs_ba)
  # inserting dummy axes for rhs_ra
  assert len(qlhs_scale_t.shape) == len(lhs_shape) - len(lhs_ca)
  start = len(qlhs_scale_t.shape)
  end = len(rhs_shape) - len(rhs_ca) - len(rhs_ba) + start
  lhs_dummy_axes = range(start, end)
  qlhs_scale_t = jnp.expand_dims(qlhs_scale_t, axis=lhs_dummy_axes)
  return qlhs_scale_t


def rhs_scale_transpose_to_output(
    rhs_scale, dimension_numbers, lhs_shape, rhs_shape
):
  """Transposes rhs_scale to output dimension order."""
  if rhs_scale is None:
    return None
  del rhs_shape
  (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = dimension_numbers
  qrhs_scale_t = _scale_trans(rhs_scale, rhs_ca, rhs_ba)
  start = len(rhs_ba)
  end = len(lhs_shape) - len(lhs_ca) - len(lhs_ba) + start
  rhs_dummy_axes = range(start, end)
  qrhs_scale_t = jnp.expand_dims(qrhs_scale_t, axis=rhs_dummy_axes)
  return qrhs_scale_t


def _scale_trans_back(
    scale_t: jax.Array, ca: Sequence[utils.AxisIdx], ba: Sequence[utils.AxisIdx]
) -> jax.Array:
  """Transposes scale (transposed for output) back to its original dimension.

  Args:
    scale_t: scale transposed for output, without other arguments' remaining
      axis dimensions. Output of _scale_trans.
    ca: contracting axis.
    ba: batching axis.

  Returns:
    Recovered scale from the scale_t.
  """
  ca, ba = list(ca), list(ba)

  start = len(scale_t.shape)
  end = start + len(ca)
  scale = jnp.expand_dims(scale_t, axis=range(start, end))

  ra = utils.get_remaining_axes(scale.ndim, ca, ba)

  transpose_back = [-1] * len(ba + ra + ca)
  for axis_orig, axis_transposed in enumerate(ba + ra + ca):
    transpose_back[axis_transposed] = axis_orig

  assert -1 not in transpose_back

  scale = transpose(scale, transpose_back)
  return scale


def lhs_recover_scale_from_scale_t(
    lhs_scale_t: jax.Array,
    dimension_numbers: jax.lax.DotDimensionNumbers,
    lhs_shape: Sequence[int],
    rhs_shape: Sequence[int],
):
  """Recovers lhs_scale from lhs_scale_t."""
  (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = dimension_numbers

  # Remove dummy axes.
  start = len(lhs_shape) - len(lhs_ca)
  rhs_ra_ndim = len(rhs_shape) - len(rhs_ca) - len(rhs_ba)
  lhs_scale = jnp.squeeze(lhs_scale_t, axis=range(start, start + rhs_ra_ndim))

  return _scale_trans_back(lhs_scale, lhs_ca, lhs_ba)


def rhs_recover_scale_from_scale_t(
    rhs_scale_t: jax.Array,
    dimension_numbers: jax.lax.DotDimensionNumbers,
    lhs_shape: Sequence[int],
    rhs_shape: Sequence[int],
):
  """Recovers rhs_scale from rhs_scale_t."""
  del rhs_shape

  (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = dimension_numbers

  start = len(rhs_ba)
  end = len(lhs_shape) - len(lhs_ca) - len(lhs_ba) + start
  rhs_scale = jnp.squeeze(rhs_scale_t, axis=range(start, end))

  return _scale_trans_back(rhs_scale, rhs_ca, rhs_ba)


def _scale_trans_for_other_input(
    x: jax.Array,
    my_ca: Sequence[utils.AxisIdx],
    my_ba: Sequence[utils.AxisIdx],
    other_ca: Sequence[utils.AxisIdx],
    other_ba: Sequence[utils.AxisIdx],
    other_rank: int,
):
  """Transposes x to other inputs' dimension order."""
  my_ca = list(my_ca)
  my_ba = list(my_ba)
  other_ca = list(other_ca)
  other_ba = list(other_ba)

  # Match the rank.
  if len(x.shape) < other_rank:
    x = x.reshape(list(x.shape) + [1] * (other_rank - len(x.shape)))

  transpose_dim = [-1] * len(x.shape)
  my_axis_mapped = my_ca + my_ba
  other_axis_mapped = other_ca + other_ba
  my_ra = utils.get_remaining_axes(x.ndim, my_ca, my_ba)

  for axis in my_ra:
    assert x.shape[axis] == 1
  for my_axis, other_axis in zip(my_axis_mapped, other_axis_mapped):
    transpose_dim[other_axis] = my_axis

  # Fill unrelated axis with remaining axis.
  ra_idx = 0
  for transpose_dim_idx, transpose_dim_value in enumerate(transpose_dim):
    if transpose_dim_value == -1:
      transpose_dim[transpose_dim_idx] = my_ra[ra_idx]
      ra_idx += 1
  assert ra_idx == len(my_ra)

  # Transpose.
  x = transpose(x, transpose_dim)

  # Remove redundant axis.
  if len(x.shape) > other_rank:
    for idx in range(len(x.shape), other_rank):
      assert x.shape[idx] == 1
    x = x.reshape(x.shape[:other_rank])

  return x


def lhs_scale_transpose_for_rhs_input(lhs_scale, dimension_numbers, rhs_shape):
  """Transposes lhs_scale to rhs input dimension order."""
  if lhs_scale is None:
    return None

  (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = dimension_numbers
  return _scale_trans_for_other_input(
      lhs_scale, lhs_ca, lhs_ba, rhs_ca, rhs_ba, len(rhs_shape)
  )


def rhs_scale_transpose_for_lhs_input(rhs_scale, dimension_numbers, lhs_shape):
  """Transposes lhs_scale to rhs input dimension order."""
  if rhs_scale is None:
    return None

  (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = dimension_numbers
  return _scale_trans_for_other_input(
      rhs_scale, rhs_ca, rhs_ba, lhs_ca, lhs_ba, len(lhs_shape)
  )



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for aqt.jax.stats_tag."""

import math

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax_legacy.jax import test_utils
from aqt.jax_legacy.jax.stats_tag import StatsTag
from jax import random
import jax.numpy as jnp
import numpy as onp

test_utils.configure_jax()


class StatsTagTest(parameterized.TestCase):

  def setUp(self):
    super(StatsTagTest, self).setUp()
    self.rng = random.PRNGKey(0)
    key1, key2 = random.split(self.rng)
    self.key = key2
    self.x = random.normal(key1, (4, 3, 2))
    self.x2 = jnp.ones((4, 3, 2))
    self.x3 = jnp.arange(-3, -3 + (4 * 3 * 2)).reshape((4, 3, 2))
    self.x4 = jnp.ones((11, 13, 15))

  def init_model(self,
                 init_x,
                 channel_axis,
                 update_stats,
                 num_indices_per_ax=10):
    tag = StatsTag(
        num_indices_per_ax=num_indices_per_ax,
        channel_axis=channel_axis,
        update_stats=update_stats)
    init_state = tag.init(self.key, init_x, mask=None)
    return tag, init_state

  def test_masking(self):
    x = jnp.array([[1, -2, 3], [4, 5, 6]]).astype(jnp.float32)
    mask = jnp.array([[False, True, True], [True, True, False]])
    model, init_state = self.init_model(
        init_x=x, channel_axis=1, update_stats=True)
    _, new_state = model.apply(init_state, x, mask=mask, mutable='stats_tag')
    stats = new_state['stats_tag']
    onp.testing.assert_allclose(stats['min_per_ch'], [4, -2, 3])
    onp.testing.assert_allclose(stats['max_per_ch'], [4, 5, 3])
    mean_channel2 = (-2 + 5) / 2
    onp.testing.assert_allclose(stats['mean_per_ch'], [4, mean_channel2, 3])
    stddev_channel2 = math.sqrt(
        ((-2 - mean_channel2)**2 + (5 - mean_channel2)**2) / 2)
    onp.testing.assert_allclose(stats['stddev_per_ch'], [0, stddev_channel2, 0])
    absdev_channel2 = (abs(-2 - mean_channel2) + abs(5 - mean_channel2)) / 2
    onp.testing.assert_allclose(stats['absdev_per_ch'], [0, absdev_channel2, 0])
    sttdev_uncentered_channel2 = math.sqrt(((-2)**2 + 5**2) / 2)
    onp.testing.assert_allclose(stats['stddev_per_ch_uncentered'],
                                [4, sttdev_uncentered_channel2, 3])
    absdev_uncentered_channel2 = (abs(-2) + abs(5)) / 2
    onp.testing.assert_allclose(stats['absdev_per_ch_uncentered'],
                                [4, absdev_uncentered_channel2, 3])

    # Test correctness in presence of broadcasting and with channel_axis=None.
    # The mask, which is 1x3, has to be broadcasted to match 'x', which is
    # 2x3. After masking and flattening, 'x' should be equivalent to
    # [1, 3, -4, 6], so we compare to statistics computed on that array.

    mask = jnp.array([True, False, True]).reshape(1, 3)
    x = jnp.array([[1, 2, 3], [-4, 5, 6]])
    model, init_state = self.init_model(
        init_x=x, channel_axis=None, update_stats=True)
    _, state = model.apply(init_state, x, mask=mask, mutable='stats_tag')
    stats = state['stats_tag']
    onp.testing.assert_allclose(stats['min_per_ch'], -4)
    onp.testing.assert_allclose(stats['max_per_ch'], 6)
    mean = (1 + 3 + -4 + 6) / 4
    onp.testing.assert_allclose(stats['mean_per_ch'], mean)
    stddev = jnp.sqrt(
        ((1 - mean)**2 + (3 - mean)**2 + (-4 - mean)**2 + (6 - mean)**2) / 4)
    onp.testing.assert_allclose(stats['stddev_per_ch'], stddev)
    stddev_uncentered = jnp.sqrt((1**2 + 3**2 + (-4)**2 + 6**2) / 4)
    onp.testing.assert_allclose(stats['stddev_per_ch_uncentered'],
                                stddev_uncentered)
    absdev = (onp.abs(1 - mean) + onp.abs(3 - mean) + onp.abs(-4 - mean) +
              onp.abs(6 - mean)) / 4
    onp.testing.assert_allclose(stats['absdev_per_ch'], absdev)
    absdev_uncentered = (onp.abs(1) + onp.abs(3) + onp.abs(-4) + onp.abs(6)) / 4
    onp.testing.assert_allclose(stats['absdev_per_ch_uncentered'],
                                absdev_uncentered)

    # Test with broadcasting and channel_axis=1
    model, init_state = self.init_model(
        init_x=x, channel_axis=(0,), update_stats=True)
    _, state = model.apply(init_state, x, mask=mask, mutable='stats_tag')
    stats = state['stats_tag']
    mean = [(1 + 3) / 2, (-4 + 6) / 2]
    onp.testing.assert_allclose(stats['mean_per_ch'], mean)
    max_per_ch = [3, 6]
    onp.testing.assert_allclose(stats['max_per_ch'], max_per_ch)

  @parameterized.named_parameters(
      dict(
          testcase_name='no_channel_dim',
          channel_axis=None,
          expected_shape=(1,)),
      dict(
          testcase_name='one_channel_dim', channel_axis=-1,
          expected_shape=(2,)),
      dict(
          testcase_name='one_channel_dim_tuple',
          channel_axis=(2,),
          expected_shape=(2,)),
      dict(
          testcase_name='two_channel_dims',
          channel_axis=(1, 2),
          expected_shape=(3, 2)),
  )
  def test_init_and_shape(self, channel_axis, expected_shape):
    _, init_state = self.init_model(
        init_x=self.x, channel_axis=channel_axis, update_stats=True)
    for key in [
        'min_per_ch', 'max_per_ch', 'mean_per_ch', 'stddev_per_ch',
        'absdev_per_ch', 'stddev_per_ch_uncentered', 'absdev_per_ch_uncentered'
    ]:
      onp.testing.assert_array_equal(init_state['stats_tag'][key],
                                     onp.zeros(expected_shape))

  @parameterized.named_parameters(
      dict(
          testcase_name='no_channel_dim',
          channel_axis=None,
          expected_shape=(1,)),
      dict(
          testcase_name='one_channel_dim', channel_axis=-1,
          expected_shape=(2,)),
      dict(
          testcase_name='two_channel_dims',
          channel_axis=(1, 2),
          expected_shape=(3, 2)),
  )
  def test_tag_stats_ones(self, channel_axis, expected_shape):
    model, init_state = self.init_model(
        init_x=self.x, channel_axis=channel_axis, update_stats=True)
    _, state_0 = model.apply(
        init_state, self.x2, mask=None, mutable='stats_tag')
    stats = state_0['stats_tag']
    for key in [
        'min_per_ch', 'max_per_ch', 'mean_per_ch', 'stddev_per_ch_uncentered',
        'absdev_per_ch_uncentered'
    ]:
      onp.testing.assert_array_equal(stats[key], onp.ones(expected_shape))
    for key in ['stddev_per_ch', 'absdev_per_ch']:
      onp.testing.assert_array_equal(stats[key], onp.zeros(expected_shape))

  @parameterized.named_parameters(
      dict(
          testcase_name='no_channel_dim',
          channel_axis=None,
          reduction_axis=(0, 1, 2)),
      dict(
          testcase_name='one_channel_dim',
          channel_axis=-1,
          reduction_axis=(0, 1)),
      dict(
          testcase_name='two_channel_dims',
          channel_axis=(1, 2),
          reduction_axis=(0,)),
  )
  def test_tag_stats_numpy_comparison(self, channel_axis, reduction_axis):
    model, init_state = self.init_model(
        init_x=self.x, channel_axis=channel_axis, update_stats=True)
    _, state_0 = model.apply(
        init_state, self.x2, mask=None, mutable='stats_tag')
    _, state_1 = model.apply(state_0, self.x3, mask=None, mutable='stats_tag')
    stats = state_1['stats_tag']
    onp.testing.assert_allclose(stats['min_per_ch'],
                                onp.min(self.x3, axis=reduction_axis))
    onp.testing.assert_allclose(stats['max_per_ch'],
                                onp.max(self.x3, axis=reduction_axis))
    onp.testing.assert_allclose(stats['mean_per_ch'],
                                onp.mean(self.x3, axis=reduction_axis))
    onp.testing.assert_allclose(stats['stddev_per_ch'],
                                onp.std(self.x3, axis=reduction_axis))
    onp.testing.assert_allclose(
        stats['absdev_per_ch'],
        onp.mean(
            onp.abs(self.x3 -
                    onp.mean(self.x3, axis=reduction_axis, keepdims=True)),
            axis=reduction_axis))
    onp.testing.assert_allclose(
        stats['stddev_per_ch_uncentered'],
        onp.sqrt(onp.mean(onp.square(self.x3), axis=reduction_axis)))
    onp.testing.assert_allclose(stats['absdev_per_ch_uncentered'],
                                onp.mean(onp.abs(self.x3), axis=reduction_axis))

  @parameterized.named_parameters(
      dict(
          testcase_name='no_channel_dim',
          channel_axis=None,
          num_indices_per_ax=5,
          expected_shape=(1,)),
      dict(
          testcase_name='one_channel_dim',
          channel_axis=-1,
          num_indices_per_ax=5,
          expected_shape=(5,)),
      dict(
          testcase_name='two_channel_dims',
          channel_axis=(1, 2),
          num_indices_per_ax=6,
          expected_shape=(6, 6)),
  )
  def test_channel_axis_sampling(self, channel_axis, num_indices_per_ax,
                                 expected_shape):
    model, init_state = self.init_model(
        init_x=self.x4,
        channel_axis=channel_axis,
        update_stats=True,
        num_indices_per_ax=num_indices_per_ax)
    _, state_0 = model.apply(
        init_state, self.x4, mask=None, mutable='stats_tag')
    onp.testing.assert_array_equal(state_0['stats_tag']['min_per_ch'],
                                   onp.ones(expected_shape))


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for aqt.jax.stats."""

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax_legacy.jax import test_utils
from aqt.jax_legacy.jax.stats import masked_mean
from aqt.jax_legacy.jax.stats import masked_sum
from aqt.jax_legacy.jax.stats import Stats
from jax import random
import jax.numpy as jnp
import numpy as onp

test_utils.configure_jax()


class StatsTest(parameterized.TestCase):

  def setUp(self):
    super(StatsTest, self).setUp()
    self.key = random.PRNGKey(0)

  def test_create_empty_stats_using_initializer(self):
    stats = Stats.stats_initializer(shape=())
    self.assertEqual(stats.n, 0)
    self.assertEqual(stats.mean, 0.)
    self.assertEqual(stats.mean_abs, 0.)
    self.assertEqual(stats.mean_sq, 0.)

  # TODO(wanglisa/shivaniagrawal): parametrize tests for values of alpha.
  @parameterized.named_parameters(
      dict(
          testcase_name='axisnone_1dim',
          axis=None,
          sample_shape=(2,),
          stats_shape=(1,),
      ),
      dict(
          testcase_name='axisnone_2dim',
          axis=None,
          sample_shape=(2, 3),
          stats_shape=(1, 1),
      ),
      dict(
          testcase_name='axisnone_3dim',
          axis=None,
          sample_shape=(2, 3, 4),
          stats_shape=(1, 1, 1),
      ),
      dict(
          testcase_name='axis0_2dim',
          axis=0,
          sample_shape=(2, 3),
          stats_shape=(1, 3),
      ),
      dict(
          testcase_name='axis0_3dim',
          axis=0,
          sample_shape=(2, 3, 4),
          stats_shape=(1, 3, 4),
      ),
      dict(
          testcase_name='axis0-1_3dim',
          axis=(0, 1),
          sample_shape=(2, 3, 4),
          stats_shape=(1, 1, 4),
      ),
  )
  def test_update_stats_with_different_axes(self, axis, sample_shape,
                                            stats_shape):
    stats = Stats.stats_initializer(shape=stats_shape)
    for i in range(-1, 4):
      x = i * jnp.ones(sample_shape)
      stats = Stats.create_updated_stats(stats, x, axis=axis)
    self.assertEqual(stats.n, 5)

    exp_mean = (-1 + 0 + 1 + 2 + 3) / 5. * jnp.ones(stats_shape)
    onp.testing.assert_allclose(stats.mean, exp_mean)

    exp_mean_abs = (1 + 0 + 1 + 2 + 3) / 5. * jnp.ones(stats_shape)
    onp.testing.assert_allclose(stats.mean_abs, exp_mean_abs)

    exp_mean_sq = (
        (-1)**2 + 0**2 + 1**2 + 2**2 + 3**2) / 5. * jnp.ones(stats_shape)
    onp.testing.assert_allclose(stats.mean_sq, exp_mean_sq)

  def test_per_channel_average(self):
    """Stats should be different per channel."""
    stats = Stats.stats_initializer(shape=())
    for i in range(-1, 4):
      x = i * jnp.array([[1., 2.], [2., 4.], [3., 6.]])
      stats = Stats.create_updated_stats(stats, x, axis=(0,))
    self.assertEqual(stats.n, 5)

    # For i in range(-1, 4), ith array would be
    # [[i    , 2 * i]
    #  [i * 2, 2 * (i * 2)]
    #. [i * 3, 2 * (i * 3)]]

    exp_mean_ch0 = (-1 + 0 + 1 + 2 + 3) * (1 + 2 + 3) / 15.
    exp_mean_ch1 = (-2 + 0 + 2 + 4 + 6) * (1 + 2 + 3) / 15.
    exp_mean = jnp.array([[exp_mean_ch0, exp_mean_ch1]])
    onp.testing.assert_allclose(stats.mean, exp_mean)

    exp_mean_abs_ch0 = (1 + 0 + 1 + 2 + 3) * (1 + 2 + 3) / 15.
    exp_mean_abs_ch1 = (2 + 0 + 2 + 4 + 6) * (1 + 2 + 3) / 15.
    exp_mean_abs = jnp.array([[exp_mean_abs_ch0, exp_mean_abs_ch1]])
    onp.testing.assert_allclose(stats.mean_abs, exp_mean_abs)

    exp_mean_sq_ch0 = (
        (-1)**2 + 0**2 + 1**2 + 2**2 + 3**2) * (1**2 + 2**2 + 3**2) / 15.
    exp_mean_sq_ch1 = (
        (-2)**2 + 0**2 + 2**2 + 4**2 + 6**2) * (1**2 + 2**2 + 3**2) / 15.
    exp_mean_sq = jnp.array([[exp_mean_sq_ch0, exp_mean_sq_ch1]])
    onp.testing.assert_allclose(stats.mean_sq, exp_mean_sq)

  def test_masking(self):
    # We will simulate a situation where we have two batches with two tokens
    # each, and the second token of the second batch is padding. The channel
    # dimension is three.

    stats = Stats.stats_initializer(shape=(1, 1, 2))
    # The shape of 'x' is [batch index, token index, channel index]
    x = jnp.reshape(jnp.arange(8), (2, 2, 2)).astype(jnp.float32)
    token_mask = jnp.array([[True, True], [True, False]])
    mask = token_mask[...,
                      None]  # Broadcast the mask over the channel dimension
    stats = Stats.create_updated_stats(stats, x, axis=(0, 1), mask=mask)
    exp_mean_ch0 = (0 + 2 + 4) / 3
    exp_mean_ch1 = (1 + 3 + 5) / 3
    exp_mean = jnp.array([[[exp_mean_ch0, exp_mean_ch1]]])
    onp.testing.assert_allclose(stats.mean, exp_mean)
    onp.testing.assert_allclose(stats.mean_abs, exp_mean)
    exp_mean_sq_ch0 = (0**2 + 2**2 + 4**2) / 3
    exp_mean_sq_ch1 = (1**2 + 3**2 + 5**2) / 3
    exp_mean_sq = jnp.array([[[exp_mean_sq_ch0, exp_mean_sq_ch1]]])
    onp.testing.assert_allclose(stats.mean_sq, exp_mean_sq)
    exp_max = [[[4, 5]]]
    onp.testing.assert_allclose(stats.mean_batch_maximum, exp_max)
    exp_min = [[[0, 1]]]
    onp.testing.assert_allclose(stats.mean_batch_minimum, exp_min)

    # Now do the same, but with axis=None
    stats = Stats.stats_initializer(shape=())
    stats = Stats.create_updated_stats(stats, x, axis=None, mask=mask)
    exp_mean = (0 + 1 + 2 + 3 + 4 + 5) / 6
    onp.testing.assert_allclose(stats.mean, [[[exp_mean]]])
    exp_mean_sq = (0**2 + 1**2 + 2**2 + 3**2 + 4**2 + 5**2) / 6
    onp.testing.assert_allclose(stats.mean_sq, [[[exp_mean_sq]]])
    onp.testing.assert_allclose(stats.mean_batch_maximum, [[[5]]])
    onp.testing.assert_allclose(stats.mean_batch_minimum, [[[0]]])

    # Also try with reduction axis equal to the broadcasting axis.
    # In this case, we expect a 0 when taking the mean over the
    # array slice that consists solely of masked elements, since only masked
    # elements will not update the initial value of 0.
    stats = Stats.stats_initializer(shape=(2, 2, 1))
    stats = Stats.create_updated_stats(stats, x, axis=(2,), mask=mask)
    exp_mean = [[[(0 + 1) / 2], [(2 + 3) / 2]], [[(4 + 5) / 2], [0]]]
    onp.testing.assert_allclose(stats.mean, exp_mean)
    exp_mean_sq = [[[(0**2 + 1**2) / 2], [(2**2 + 3**2) / 2]],
                   [[(4**2 + 5**2) / 2], [0]]]
    onp.testing.assert_allclose(stats.mean_sq, exp_mean_sq)
    exp_max = [[[1], [3]], [[5], [0]]]
    onp.testing.assert_allclose(stats.mean_batch_maximum, exp_max)
    exp_min = [[[0], [2]], [[4], [0]]]
    onp.testing.assert_allclose(stats.mean_batch_minimum, exp_min)

  @parameterized.named_parameters(
      dict(
          testcase_name='axis_none',
          axis=None,
          exp_masked_sum=(0 + 1 + 2 + 3 + 4 + 5),
          exp_masked_mean=(0 + 1 + 2 + 3 + 4 + 5) / 6,
      ),
      dict(
          testcase_name='axis_01',
          axis=(0, 1),
          exp_masked_sum=onp.array([0 + 2 + 4, 1 + 3 + 5]),
          exp_masked_mean=onp.array([(0 + 2 + 4) / 3, (1 + 3 + 5) / 3]),
      ),
      dict(
          testcase_name='axis_0',
          axis=(0,),
          exp_masked_sum=onp.array([[0 + 4, 1 + 5], [2 + 0, 3 + 0]]),
          exp_masked_mean=onp.array([[(0 + 4) / 2, (1 + 5) / 2],
                                     [(2 + 0) / 1, (3 + 0) / 1]]),
      ),
  )
  def test_masked_sum_and_masked_mean(self, axis, exp_masked_sum,
                                      exp_masked_mean):
    x = jnp.reshape(jnp.arange(8), (2, 2, 2)).astype(jnp.float32)
    mask = jnp.array([[[True, True], [True, True]],
                      [[True, True], [False, False]]])
    masked_sum_res = masked_sum(
        x, mask=mask, axis=axis, paxis_name=None, keepdims=False)
    onp.testing.assert_allclose(masked_sum_res, exp_masked_sum)

    masked_mean_res = masked_mean(
        x, mask=mask, axis=axis, paxis_name=None, keepdims=False)
    onp.testing.assert_allclose(masked_mean_res, exp_masked_mean)

  @parameterized.named_parameters(
      dict(
          testcase_name='axis_none_no_mask',
          x=onp.reshape(onp.arange(4), (2, 2)).astype(onp.float32),
          mask=None,
          axis=None,
          exp_mean=[[(1 + 2 + 3) / 3]],
      ),
      dict(
          testcase_name='axis_none_with_mask',
          x=onp.reshape(onp.arange(8), (2, 2, 2)).astype(onp.float32),
          mask=onp.array([[[True, True], [True, True]],
                          [[True, True], [False, False]]]),
          axis=None,
          exp_mean=[[[(1 + 2 + 3 + 4 + 5) / 5]]],
      ),
      # Because mean of x will have div by 0, it's an invalid mean, so
      # it should not update the initial value of 0.
      dict(
          testcase_name='div_by_0_axis_none_no_mask',
          x=onp.array([0, 0]).astype(onp.float32),
          mask=None,
          axis=None,
          exp_mean=[0],
      ),
      dict(
          testcase_name='div_by_0_axis_none_with_mask',
          x=onp.array([0, 1]).astype(onp.float32),
          mask=onp.array([True, False]),
          axis=None,
          exp_mean=[0],
      ),
      dict(
          testcase_name='2dim_div_by_0_axis_0_no_mask',
          x=onp.array([[0, 1], [0, 0]]).astype(onp.float32),
          mask=None,
          axis=(0,),
          exp_mean=onp.array([[0, (1 + 0) / 1]]),
      ),
      # Because mean[0] of x will have div by 0, it's an invalid mean, so
      # it should not update the initial value of 0.
      dict(
          testcase_name='2dim_div_by_0_axis_0_with_mask',
          x=onp.array([[0, 1], [2, 3]]).astype(onp.float32),
          mask=onp.array([[True, True], [False, True]]),
          axis=(0,),
          exp_mean=onp.array([[0, (1 + 3) / 2]]),
      ),
  )
  def test_exclude_zeros(self, x, mask, axis, exp_mean):
    """Stats should be different when excluding zeros."""
    stats = Stats.stats_initializer(shape=())
    stats = Stats.create_updated_stats(
        stats, x, axis=axis, mask=mask, exclude_zeros=True)
    onp.testing.assert_allclose(stats.mean, exp_mean)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax.v2 import config
from aqt.jax.v2 import utils
from aqt.jax.v2.examples import flax_e2e_model
from aqt.jax.v2.numerics import fp8_numerics
import jax
import jax.numpy as jnp
import numpy as np


def averaged_stochastic_rounding(
    numerics,
    key,
    x_count,
    x_min,
    x_max,
    sr_count,
):
  x = jnp.linspace(x_min, x_max, x_count, dtype=jnp.float32)
  bx = x.reshape((x_count, 1)).astype(jnp.bfloat16)
  bx = jnp.ones((x_count, sr_count), dtype=jnp.bfloat16) * bx  # broadcast
  context = utils.Context(key=jax.random.PRNGKey(key), train_step=None)
  qx, _ = numerics.vjp_fwd(bx, context)
  assert qx.dtype == numerics.get_dtype()
  qx = qx.astype(jnp.float32)
  qx = jnp.mean(qx, axis=1)
  return x, qx


class MyTest(parameterized.TestCase):

  def test_2bit_mantissa(self):
    x = jnp.array(
        [
            [0, 1, -1, 0.1234567, 100, 0.5, 0.00000000001, 0.01],
            [
                -2.9098532,
                -0.8508397,
                -0.14903745,
                -3.572757,
                0.63277215,
                0.7478278,
                1.3707844,
                1.3319879,
            ],
        ],
        dtype=jnp.float32,
    )

    result = fp8_numerics.round_to_nearest_even(x, jnp.float8_e5m2)

    np.testing.assert_array_equal(
        result,
        np.array(
            [
                [0, 1, -1, 0.125, 96, 0.5, 0, 10 / 1024],
                [-3, -0.875, -0.15625, -3.5, 0.625, 0.75, 1.25, 1.25],
            ],
            dtype=jnp.float32,
        ),
    )

  def test_3bit_mantissa(self):
    x = jnp.array(
        [
            [0, 1, -1, 0.1234567, 100, 0.5, 0.00000000001, 0.01],
            [
                -2.9098532,
                -0.8508397,
                -0.14903745,
                -3.572757,
                0.63277215,
                0.7478278,
                1.3707844,
                1.3319879,
            ],
        ],
        dtype=jnp.float32,
    )

    result = fp8_numerics.round_to_nearest_even(x, jnp.float8_e4m3fn)

    np.testing.assert_array_equal(
        result,
        np.array(
            [
                [0, 1, -1, 0.125, 96, 0.5, 0, 10 / 1024],
                [-3, -0.875, -0.15625, -3.5, 0.625, 0.75, 1.375, 1.375],
            ],
            dtype=jnp.float32,
        ),
    )

  def test_retains_dtype(self):
    result_f16 = fp8_numerics.round_to_nearest_even(
        jnp.array([1], dtype=jnp.bfloat16), jnp.float8_e5m2
    )
    result_f32 = fp8_numerics.round_to_nearest_even(
        jnp.array([1], dtype=jnp.float32), jnp.float8_e4m3fn
    )

    self.assertEqual(result_f32.dtype, jnp.float8_e4m3fn)
    self.assertEqual(result_f16.dtype, jnp.float8_e5m2)

  @parameterized.parameters([
      dict(fwd_bits="e4m3"),
      dict(fwd_bits="e5m2"),
  ])
  def test_mnist_training(self, fwd_bits: str):
    target_loss = {
        "e4m3": {
            "cpu": [
                # Different CPU models are not bit exact and sometimes produce
                # different losses under the same training setting.
                3.167573690414428710937500000000,
                3.167574644088745117187500000000,
                3.167066574096679687500000000000,  # colab
            ],
            "TPU v2": [3.2136783599853515625],
            "TPU v3": [3.2136783599853515625],
            "TPU v4": [3.2137317657470703125],
            "TPU v5 lite": [3.21373653411865234375],
        },
        "e5m2": {
            "cpu": [
                # Different CPU models are not bit exact and sometimes produce
                # different losses under the same training setting.
                3.112026214599609375000000000000,
                3.112027645111083984375000000000,
                3.112027168273925781250000000000,  # colab s
            ],
            "TPU v2": [3.2267177104949951171875],
            "TPU v3": [3.2267177104949951171875],
            "TPU v4": [3.2266914844512939453125],
            "TPU v5 lite": [3.2268390655517578125],
        },
    }

    # RNGs
    rng = jax.random.key(0)
    rng, init_rng = jax.random.split(rng)
    rng, image_rng = jax.random.split(rng)
    rng, label_rng = jax.random.split(rng)
    rng, input_rng = jax.random.split(rng)
    del rng

    # Dataset
    ds_size = 8
    ds = {
        "image": jax.random.uniform(key=image_rng, shape=(ds_size, 28, 28, 1)),
        "label": jax.random.randint(
            key=label_rng, shape=(ds_size,), minval=0, maxval=10
        ),
    }

    aqt_cfg = config.config_fwd_fp8(fwd_bits)
    state = flax_e2e_model.create_train_state(init_rng, aqt_cfg)

    _, train_loss, _ = flax_e2e_model.train_epoch(
        state, ds, batch_size=ds_size // 2, rng=input_rng
    )

    device = jax.devices()[0].device_kind
    if train_loss not in target_loss[fwd_bits][device]:
      msg = f"train_loss={train_loss:.30f}, {device=}, {fwd_bits=}"
      self.fail(msg)

  @parameterized.parameters([
      dict(),
      dict(key=1),
      dict(key=2),
      dict(key=3),
      dict(key=4),
      dict(dtype=jnp.float8_e4m3fn, exponent_bits=4, mantissa_bits=3),
      dict(x_min=64 / 2**20, x_max=2 * 64 / 2**20),
      dict(x_min=32 / 2**20, x_max=2 * 32 / 2**20),
      # dict(x_min=16 / 2**20, x_max=2 * 16 / 2**20),  # This one is failing.
  ])
  def test_fp8_stochastic_rounding(
      self,
      key=0,
      x_count=1024,
      x_min=1.0,
      x_max=4.0,
      sr_count=10000,
      dtype=jnp.float8_e5m2,
      exponent_bits=5,
      mantissa_bits=2,
  ):
    numerics = fp8_numerics.Fp8Numerics(
        dtype=dtype,
        exponent_bits=exponent_bits,
        mantissa_bits=mantissa_bits,
        stochastic_rounding=True,
    )
    x, qx = averaged_stochastic_rounding(
        numerics,
        key=key,
        x_count=x_count,
        x_min=x_min,
        x_max=x_max,
        sr_count=sr_count,
    )
    err = x - qx
    mean_err = jnp.mean(err)
    std_err = jnp.std(err) / jnp.sqrt(sr_count)
    mean_in_std_err_units = jnp.abs(mean_err) / std_err
    assert (
        mean_in_std_err_units < 5
    ), f"mean_in_std_err_units: {mean_in_std_err_units}"


def illustrate_bf16():
  def bit1(fro, to):
    # (fro-to) 1s followed by (to) 0s
    return (1 << fro) - (1 << to)

  for i in range(-1, 16):
    bits = jnp.uint16(0)
    if i > 0:
      bits += jnp.uint16(1 << i)
    bits += bit1(15, 8)
    f = float(jax.lax.bitcast_convert_type(bits, jnp.bfloat16))
    print(f"{bits:016b} {i=: 3}, {f:.9f}")


def illustrate_bf16_2():
  def pr(n):
    n = jnp.int16(n)
    bx = jax.lax.bitcast_convert_type(n, jnp.bfloat16)
    # fp8 = fp8_numerics.fp_mantissa_round(bx, 2, key=jax.random.PRNGKey(0))
    # fp8 = bx.astype(jnp.float8_e5m2)
    # fp8_n = jax.lax.bitcast_convert_type(fp8, jnp.uint8)
    ns = f"{n:016b}"
    ns = ns[:1] + "s " + ns[1:9] + "e " + ns[9:] + "m"
    # print("x1", fp8, fp8_n)
    # ns_8 = f"{fp8_n:08b}"
    # ns_8 = ns_8[:1] + "s " + ns_8[1:6] + "e " + ns_8[6:] + "m"
    # print(f"{float(bx):03.8f}[{ns}] -> {float(fp8):03.8}[{ns_8}]")
    print(f"{float(bx):03.8f}[{ns}]")

  pr(0b0_01111111_0000000)
  pr(0b0_01111111_0001000)
  pr(0b0_01111111_0010000)  #
  pr(0b0_01111111_0011000)
  print()

  pr(0b0_01111111_0100000)
  pr(0b0_01111111_0101000)
  pr(0b0_01111111_0110000)  #
  pr(0b0_01111111_0111000)
  print()

  pr(0b0_01111111_1000000)
  pr(0b0_01111111_1001000)
  pr(0b0_01111111_1010000)  #
  pr(0b0_01111111_1011000)
  print()

  pr(0b0_01111111_1100000)
  pr(0b0_01111111_1101000)
  pr(0b0_01111111_1110000)  #
  pr(0b0_01111111_1111000)


# To be used in colab.
def plot_sr_error(x_min=1.0, x_max=2.0, sr_count=64 * 1024, x_count=1024):
  n = fp8_numerics.Fp8Numerics(
      dtype=jnp.float8_e5m2,
      exponent_bits=5,
      mantissa_bits=2,
      stochastic_rounding=True,
  )
  x_min, x_max = 1.0, 2.0
  x, qx = averaged_stochastic_rounding(
      n, key=0, x_min=x_min, x_max=x_max, sr_count=sr_count, x_count=x_count
  )

  import matplotlib.pyplot as plt  # pylint: disable=g-import-not-at-top

  plt.figure(figsize=(30, 15))
  err = qx - x.astype(jnp.bfloat16).astype(jnp.float32)
  plt.plot(x, err)
  print(jnp.std(err), jnp.std(err) - 0.0004826417)

  plt.show()


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Static range calibration."""

from typing import Sequence

from aqt.jax.v2 import calibration
from aqt.jax.v2 import utils
from aqt.jax.v2.numerics import numerics
import flax.linen as nn
from jax import numpy as jnp


_SUM_OF_ONES = "sum_of_ones"
_SUM_OF_VALS = "sum_of_vals"
_MAX_OF_ABS_VALS = "max_of_abs_vals"
_SUM_OF_L1_VALS = "sum_of_l1_vals"
_SUM_OF_LP_VALS = "sum_of_lp_vals"


@utils.flax_slots_kw_only_dataclass
class MeanOfAbsMaxCalibration(calibration.Calibration, nn.Module):
  """State for static range AQT/PTQ/QAT."""

  quant_collection: str

  @nn.compact
  def get_bound(
      self,
      x: jnp.ndarray,
      shared_axes: None | Sequence[utils.AxisIdx],
      context: None | utils.Context = None,
  ) -> jnp.ndarray:
    abs_max = jnp.max(jnp.abs(x), axis=shared_axes, keepdims=True)
    quant_mode = context.quant_mode if context else utils.QuantMode.TRAIN

    # Initialize the max_stat.value as 0; does not update the max_stat value
    # during initialization.
    is_initializing = not self.has_variable(self.quant_collection, "sum_of_max")

    sum_of_max = self.variable(
        self.quant_collection,
        "sum_of_max",
        jnp.zeros,
        abs_max.shape,
        abs_max.dtype,
    )
    count = self.variable(
        self.quant_collection, "count", jnp.zeros, (), jnp.int32
    )

    if quant_mode == utils.QuantMode.CALIBRATE and not is_initializing:
      sum_of_max.value = sum_of_max.value + abs_max
      count.value = count.value + 1

    # TODO(dhchoi): Introduce checker for count.value.
    # This assertion should be done on jitted graph, and this is not
    # straightforward (or does not have a good option).
    # One option is jax.experimental.checkify
    # (https://jax.readthedocs.io/en/latest/debugging/checkify_guide.html), but
    # we need to transform the whole function with jax.experimental.checkify to
    # use this feature.
    # Maybe wait for the JAX language upgrade to have a better support for this?
    return sum_of_max.value / count.value

  def get_scale_and_bias(
      self,
      x: jnp.ndarray,
      shared_axes: None | Sequence[utils.AxisIdx],
      numerics_: numerics.AqtNumerics,
      context: None | utils.Context = None,
  ) -> tuple[list[jnp.ndarray], list[jnp.ndarray]]:
    dtype = self.dtype if self.dtype is not None else x.dtype
    bound = self.get_bound(x, shared_axes, context)
    scale = bound / numerics_.get_quant_bound()
    scale = calibration.ceil_to_po2(scale) if self.po2_scale else scale
    return [scale.astype(dtype)], []


# TODO: b/335764538 - Check the math correctness of the module.
class WeightedStatsCalibration(calibration.Calibration, nn.Module):
  """Migration of AQTv1 calibration to AQTv2."""

  quant_collection: str

  # Calibration coefficients
  l1_dev_coeff: float = 0.0
  lp_dev_coeff: float = 0.0
  max_dev_coeff: float = 0.0
  const_bound_coeff: float = 0.0

  # Priors
  update_count_prior: float = 1.0
  max_dev_prior: float = 0.0
  mean_prior: float = 0.0
  l1_dev_prior: float = 0.0
  lp_dev_prior: float = 0.0

  lp_order: int = 2
  safe_divide: bool = False
  filter_zeros: bool = True
  tpu_cross_replica_sum: bool = False
  ema_update_count: float = 100

  def _get_value(self, name: str):
    return self.get_variable(self.quant_collection, name)

  def _mean(self) -> jnp.ndarray:
    sum_of_vals = self._get_value(_SUM_OF_VALS)
    sum_of_ones = self._get_value(_SUM_OF_ONES)
    return self._divide(sum_of_vals, sum_of_ones)

  def _max_dev(self) -> jnp.ndarray:
    return self._get_value(_MAX_OF_ABS_VALS)

  def _l1_dev(self) -> jnp.ndarray:
    sum_of_l1_vals = self._get_value(_SUM_OF_L1_VALS)
    sum_of_ones = self._get_value(_SUM_OF_ONES)
    return self._divide(sum_of_l1_vals, sum_of_ones)

  def _lp_dev(self) -> jnp.ndarray:
    sum_of_lp_vals = self._get_value(_SUM_OF_LP_VALS)
    sum_of_ones = self._get_value(_SUM_OF_ONES)
    if self.lp_order == 2:
      # sqrt() is numerically more accurate
      return jnp.sqrt(self._divide(sum_of_lp_vals, sum_of_ones))
    else:
      # TODO(b/205769820): Make sure if the output of pow op below is
      # numerically valid.
      return self._divide(sum_of_lp_vals, sum_of_ones) ** (1.0 / self.lp_order)

  def _update_var(
      self,
      var: nn.Variable,
      s: jnp.ndarray,
      shared_axes: None | Sequence[utils.AxisIdx],
      weight: None | jnp.ndarray = None,
      reduce_max: bool = False,
  ):
    """Updates the given Flax variable."""
    if weight is not None:
      if reduce_max:
        # A maximum is a maximum, regardless of its nonzero weight.
        s = jnp.where(weight > 0, s, 0)
      else:
        s = s * weight

    reduce_fn = jnp.max if reduce_max else jnp.sum
    s = reduce_fn(s, axis=shared_axes, keepdims=True)
    if self.tpu_cross_replica_sum:
      raise NotImplementedError(
          "support for tpu_cross_replica_sum=True is not implemented"
      )
    rate = 1.0 / self.ema_update_count
    var.value = (1.0 - rate) * var.value + rate * s

  def _divide(self, x: jnp.ndarray, y: jnp.ndarray) -> jnp.ndarray:
    if self.safe_divide:
      # Safe-divide without division-by-zero.
      res = jnp.where(y != 0, x / y, jnp.zeros_like(x))
      return res

    return jnp.divide(x, y)

  @nn.compact
  def get_bound(
      self,
      x: jnp.ndarray,
      shared_axes: None | Sequence[utils.AxisIdx],
      context: None | utils.Context = None,
  ) -> jnp.ndarray:
    if self.lp_order > 30:
      raise NotImplementedError("For higher norms we should add stabilization.")

    quant_mode = context.quant_mode if context else utils.QuantMode.TRAIN
    stats_shape = list(x.shape)
    if shared_axes is not None:
      for axis in shared_axes:
        stats_shape[axis] = 1

    def _make_var(name: str, init_val: float) -> nn.Variable:
      def init_var_fn(init_val: float) -> jnp.ndarray:
        return jnp.full(stats_shape, init_val, dtype=jnp.float32)

      return self.variable(self.quant_collection, name, init_var_fn, init_val)

    sum_of_ones = _make_var(_SUM_OF_ONES, self.update_count_prior)
    sum_of_vals = _make_var(
        _SUM_OF_VALS, self.mean_prior * self.update_count_prior
    )
    max_of_abs_vals = _make_var(_MAX_OF_ABS_VALS, self.max_dev_prior)
    sum_of_l1_vals = _make_var(
        _SUM_OF_L1_VALS, self.l1_dev_prior * self.update_count_prior
    )
    sum_of_lp_vals = _make_var(
        _SUM_OF_LP_VALS,
        self.lp_dev_prior**self.lp_order * self.update_count_prior,
    )

    if quant_mode in [utils.QuantMode.TRAIN, utils.QuantMode.CALIBRATE]:
      # Update variables.

      # Layers such as ReLU emit zeros often. In such cases, we can model
      # the non-sparse distribution of weights separately, resulting in
      # unbiased estimation of non-sparse mean l1 and lp.
      # This clips away less of the distribution of inputs.
      if self.filter_zeros:
        ones = jnp.where(x != 0, 1.0, 0.0)
      else:
        ones = jnp.ones_like(x)

      px = x if self.lp_order % 2 == 0 else jnp.abs(x)

      self._update_var(sum_of_ones, ones, shared_axes)
      self._update_var(sum_of_vals, x, shared_axes)
      self._update_var(
          max_of_abs_vals, jnp.abs(x), shared_axes, reduce_max=True
      )
      self._update_var(sum_of_l1_vals, jnp.abs(x), shared_axes)
      self._update_var(sum_of_lp_vals, px**self.lp_order, shared_axes)

    return (
        self.l1_dev_coeff * self._l1_dev()
        + self.lp_dev_coeff * self._lp_dev()
        + self.max_dev_coeff * self._max_dev()
        + self.const_bound_coeff
    )

  def get_scale_and_bias(
      self,
      x: jnp.ndarray,
      shared_axes: None | Sequence[utils.AxisIdx],
      numerics_: numerics.AqtNumerics,
      context: None | utils.Context = None,
  ) -> tuple[list[jnp.ndarray], list[jnp.ndarray]]:
    dtype = self.dtype if self.dtype is not None else x.dtype
    bound = self.get_bound(x, shared_axes, context)
    scale = bound / numerics_.get_quant_bound()
    scale = calibration.ceil_to_po2(scale) if self.po2_scale else scale
    return [scale.astype(dtype)], []



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Convenience script to generate commonly used hparams configs."""

# TODO(b/175245107): Remove this file.

import dataclasses
import enum
import pathlib
import re
import time
import typing
from typing import Optional, Type, TypeVar

from absl import flags

from aqt.jax_legacy.jax import flax_attention as aqt_flax_attention
from aqt.jax_legacy.jax import flax_layers as aqt_flax_layers
from aqt.jax_legacy.jax import quant_config
from aqt.jax_legacy.jax.flax import struct as flax_struct
from aqt.jax_legacy.jax.quantization import ActsBoundT
from aqt.jax_legacy.jax.quantization import QuantOps
from aqt.jax_legacy.jax.quantization import QuantType
from aqt.jax_legacy.jax.wmt_mlperf import models
from aqt.jax_legacy.jax.wmt_mlperf import training_hparams
from aqt.jax_legacy.utils import hparams_utils as os_hparams_utils

T = TypeVar('T')


# Default fixed bound and getbounds coefficients
# These values were determined empirically to be reasonable defaults.

DEFAULTS = {}
DEFAULTS['mlp_dense_1'] = {
    'fixed': 1.0,
    'absdev': 2.5,
    'stddev': 2.5,
    'mix': 1.0
}

DEFAULTS['mlp_dense_2'] = {
    'fixed': 6.0,
    'absdev': 13.0,
    'stddev': 13.0,
    'mix': 1.0
}

DEFAULTS['att_dense_kqv'] = {
    'fixed': 1.0,
    'absdev': 4.0,
    'stddev': 5.0,
    'mix': 0.0
}

DEFAULTS['att_dense_out'] = {
    'fixed': 2.0,
    'absdev': 0.1,
    'stddev': 0.3,
    'mix': 0.0
}

DEFAULTS['logits'] = {  #
    'fixed': 1.0,
    'absdev': 3.0,
    'stddev': 3.0,
    'mix': 0.0
}

DEFAULTS['attn_act_k'] = {
    'fixed': 1.0,
    'absdev': 2.0,
    'stddev': 1.0,
    'mix': 0.0
}

DEFAULTS['attn_act_q'] = {
    'fixed': 1.0,
    'absdev': 2.0,
    'stddev': 2.0,
    'mix': 0.0
}

DEFAULTS['attn_act_v'] = {
    'fixed': 6.0,
    'absdev': 6.0,
    'stddev': 5.5,
    'mix': 0.0
}

flags.DEFINE_string(
    'filename',
    default=None,
    help='Where to store the created hparams json file.')


# Base configuration flags
class BaseConfigSize(enum.Enum):
  """Specifies the size of the transformer."""

  MINIMAL_MODEL = enum.auto()
  SMALL_MODEL = enum.auto()
  FULL_MODEL = enum.auto()


# TODO(malmaud): Get rid of these flags that implicitly define a base config
# hparams instance (specifically, base_config_size, base_config_quant_target,
# and base_config_prec). Instead, only hparams_config_filename will be used.
flags.DEFINE_enum_class(
    'base_config_size',
    default=BaseConfigSize.SMALL_MODEL,
    enum_class=BaseConfigSize,
    help=(
        'Model size of the base configuration. '
        'FULL_MODEL creates a 6-layer model. '
        'SMALL_MODEL creates a smaller 2-layer model with channel dimensions '
        'and number of heads reduced by half compared to the full model.'
        'MINIMAL_MODEL creates a 1-layer model with minimal dimensions for '
        'testing. See training_hparams.create_from_base_config() for exact values'
        ' for each model size.'))


class BaseConfigQuantTarget(enum.Enum):
  """Specifies which part of the model to quantize."""

  NONE = enum.auto()
  WEIGHTS_ONLY = enum.auto()  # only weights get quantized
  WEIGHTS_AND_SOME_ACTS_1 = enum.auto()  # All matmuls in attention and MLP
  # blocks get quantized, except act-act matmuls i.e.: K*Q and softmax*V
  # fixed bound is used for activation quantization.
  WEIGHTS_AND_SOME_ACTS_2 = enum.auto()  # This is WEIGHTS_AND_SOME_ACTS_1 and
  # additionaly quantizes logits inputs for logits_via_embedding.
  WEIGHTS_AND_ACTS = enum.auto()  # This quantizes all the weights and
  # activations for matmul.


flags.DEFINE_enum_class(
    'base_config_quant_target',
    default=BaseConfigQuantTarget.NONE,
    enum_class=BaseConfigQuantTarget,
    help=(
        'Which aspects of the model to quantize. '
        'NONE creates a model with no quantization. '
        'WEIGHTS_ONLY creates a model where all weights in all layers are quantized. '
        'WEIGHTS_AND_SOME_ACTS_1 creates a model where all weights are quantized, and most '
        'activations are quantized (the MLP dense layers and the KQV and output '
        'operations within the attention layers.'
        'WEIGHTS_AND_SOME_ACTS_2 creates a "WEIGHTS_AND_SOME_ACTS_1" model and '
        'additionally quantizes logits inputs for logits_via_embedding.'
        'WEIGHTS_AND_ACTS creates a model where all weights and activations are'
        'quantized for matmul.'))

flags.DEFINE_string(
    'base_config_prec',
    default='-1',
    help=('Precision of the quantization. Could either be specified as an '
          '(1) integer, (2) a string describing a reduced precision '
          'floating-point type. Applies to all weights and '
          'to the activations specified by the base_config_quant_target flag.'
          'If base_config_quant_target is BaseConfigQuantTarget.NONE, please'
          ' set this flag to -1. Details of formats (1) and (2): '
          '(1) Number of bits to use for quantization. (2) A string of the '
          'form "fp_quant:scaled=<true|false>,exp_min=<integer>,exp_max='
          '<integer>,sig_bits=<integer>".'))


@dataclasses.dataclass
class BaseConfig:
  """Base config parameters."""
  size: BaseConfigSize
  prec: QuantOps.PrecT
  quant_target: BaseConfigQuantTarget

  @classmethod
  def create_from_flags(cls: Type[T]) -> T:
    return cls(
        size=FLAGS.base_config_size,
        prec=parse_base_config_prec(FLAGS.base_config_prec),
        quant_target=FLAGS.base_config_quant_target)


class ConfigGenerator(enum.Enum):
  """Specifies which configs to generate."""
  BFLOAT16 = enum.auto()  # will generate configs for bfloat16 models in three
  # model sizes.
  WEIGHTS_ONLY = enum.auto()
  WEIGHTS_AND_FIXED_ACTS = enum.auto()
  WEIGHTS_AND_AUTO_ACTS = enum.auto()


flags.DEFINE_enum_class(
    'config_generator',
    default=None,
    enum_class=ConfigGenerator,
    help='Name of config generation method to call.'
    'BFLOAT16 creates the bfloat16 configs in three model sizes.')


def parse_base_config_prec(prec: Optional[str]) -> QuantOps.PrecT:
  """Parse FLAGS.base_config_prec flag value.

  Examples prec values:
    - '-1' : Special value -1, is documented in FLAGS.base_config_prec and
      corresponds to (prec is None).
    - '4'
    - '8'
    - 'fp_quant:scaled=true,exp_min=-11,exp_max=4,sig_bits=3'

  Args:
    prec: a string representing the requested precision.

  Returns:
    Parsed string converted to a QuantOps.PrecT.
  """
  if prec is None:
    return None
  try:
    prec_val = int(prec)
    return None if prec_val == -1 else prec_val
  except ValueError as valueerror:
    fp_quant_re = re.compile(
        r"""
    fp_quant:
    scaled=(true|false),
    exp_min=(-?[0-9]+),
    exp_max=(-?[0-9]+),
    sig_bits=([0-9]+)""", re.X)
    match_value = fp_quant_re.match(prec)
    if not match_value:
      raise ValueError(f'Failed to parse precision {prec}.') from valueerror
    return QuantOps.FloatQuant(
        is_scaled=match_value.group(1) == 'true',
        fp_spec=QuantOps.FloatQuant.FloatPrec(
            exp_min=int(match_value.group(2)),
            exp_max=int(match_value.group(3)),
            sig_bits=int(match_value.group(4))))


FLAGS = flags.FLAGS

dataclass = flax_struct.dataclass if not typing.TYPE_CHECKING else dataclasses.dataclass


# TODO(wanglisa): Add additional parameters to this that are still currently
# defined in the transformer_kwargs dict in train.py.
def create_base_transformer_hparams(
    *,  #
    mlp_weight_prec: QuantOps.PrecT,
    embedding_weight_prec: QuantOps.PrecT,
    attention_weight_prec: QuantOps.PrecT,
    mlp_pos_inputs_prec: QuantOps.PrecT,
    mlp_pos_inputs_hyper: ActsBoundT,
    mlp_signed_inputs_prec: QuantOps.PrecT,
    mlp_signed_inputs_hyper: ActsBoundT,
    attention_kqv_inputs_prec: QuantOps.PrecT,
    attention_kqv_inputs_hyper: ActsBoundT,
    attention_out_inputs_prec: QuantOps.PrecT,
    attention_out_inputs_hyper: ActsBoundT,
    attention_act_q_inputs_prec: QuantOps.PrecT,
    attention_act_q_inputs_hyper: ActsBoundT,
    attention_act_k_inputs_prec: QuantOps.PrecT,
    attention_act_k_inputs_hyper: ActsBoundT,
    attention_act_probs_inputs_prec: QuantOps.PrecT,
    attention_act_v_inputs_prec: QuantOps.PrecT,
    attention_act_v_inputs_hyper: ActsBoundT,
    logits_inputs_prec: QuantOps.PrecT,
    logits_inputs_hyper: ActsBoundT,
    logits_via_embeddings: bool,
    num_layers: int,
    num_heads: int,
    emb_dim: int,
    qkv_dim: int,
    mlp_dim: int,
    quant_type: QuantType,
    half_shift: bool = False,
) -> models.Transformer.HParams:
  """Configure a transformer based on high-level configuration parameters.

  The encoder and decoder share hyperparameters.

  Args:
    mlp_weight_prec: Weight precision used by the MLP layers. `None` indicates
      no quantization.
    embedding_weight_prec: Weight precision used for weights in the embedding
      layers. `None` indicates no quantization.
    attention_weight_prec: Weight precision used for weights in the attention
      layers. `None` indicates no quantization.
    mlp_pos_inputs_prec: Positive input precision for the dense layers in the
      encoder and decoder.
    mlp_pos_inputs_hyper: Activation hyperparam for positive input quantization
      for the dense layers in the encoder and decoder.
    mlp_signed_inputs_prec: Signed input precision for the dense layers in the
      encoder and decoder.
    mlp_signed_inputs_hyper: Activation hyperparam for signed input quantization
      for the dense layers in the encoder and decoder.
    attention_kqv_inputs_prec: Input precision for the attention k in the
      encoder and decoder.
    attention_kqv_inputs_hyper: Activation hyperparam for input quantization for
      the attention k in the encoder and decoder.
    attention_out_inputs_prec: Input precision for the attention out in the
      encoder and decoder.
    attention_out_inputs_hyper: Activation hyperparam for input quantization for
      the attention out in the encoder and decoder.
    attention_act_q_inputs_prec: Input precision for the attention query(Q) for
      Q*K in the encoder and decoder.
    attention_act_q_inputs_hyper: Activation hyperparam for input query(Q)
      quantization for the attention Q*K in the encoder and decoder.
    attention_act_k_inputs_prec: Input precision for the attention key(K) for
      Q*K in the encoder and decoder.
    attention_act_k_inputs_hyper: Activation hyperparam for input key(K)
      quantization for the attention Q*K in the encoder and decoder.
    attention_act_probs_inputs_prec: Input precision for the attention weights
      i.e. Softmax(Q*K) for Softmax(Q*K)*V  in the encoder and decoder.
    attention_act_v_inputs_prec: Input precision for the attention value(V) for
      Softmax(Q*K)*V in the encoder and decoder.
    attention_act_v_inputs_hyper: Activation hyperparam for input value(V)
      quantization for the attention Softmax(Q*K)*V in the encoder and decoder.
    logits_inputs_prec: Input precision for the logits in decoder.
    logits_inputs_hyper: Activation hyperparam for input quantization for logits
      in decoder.
    logits_via_embeddings: Whether to compute logits using the embeddings
      weights.
    num_layers: Number of layers in the encoder and decoder.
    num_heads: number of heads.
    emb_dim: dimension of embedding.
    qkv_dim: dimension of the query/key/value.
    mlp_dim: dimension of the mlp on top of attention block.
    quant_type: quantization strategy for quantizing the model, one of
      `fake_quant` or `aqt`.
    half_shift: Half_shift flag used for both weights and inputs in the model.

  Returns:
    An instance of models.Transformer.HParams that contains a complete
      configuration of an encoder-decoder transformer.
  """
  embedding = aqt_flax_layers.EmbedAqt.HParams(
      weight_prec=embedding_weight_prec,
      weight_half_shift=half_shift,
      quant_act=QuantOps.ActHParams(
          input_distribution=QuantOps.ActHParams.InputDistribution.SYMMETRIC,
          prec=logits_inputs_prec,
          half_shift=half_shift,
          bounds=logits_inputs_hyper),
      quant_type=quant_type)
  # TODO(malmaud): Add flags to control weight_quant_granularity here and below.
  attention_kqv_dense = aqt_flax_layers.DenseAqt.HParams(
      weight_prec=attention_weight_prec,
      weight_half_shift=half_shift,
      quant_act=QuantOps.ActHParams(
          input_distribution=QuantOps.ActHParams.InputDistribution.SYMMETRIC,
          prec=attention_kqv_inputs_prec,
          half_shift=half_shift,
          bounds=attention_kqv_inputs_hyper),
      quant_type=quant_type,
      weight_quant_granularity=quant_config.QuantGranularity.PER_CHANNEL)
  attention_out_dense = aqt_flax_layers.DenseAqt.HParams(
      weight_prec=attention_weight_prec,
      weight_half_shift=half_shift,
      quant_act=QuantOps.ActHParams(
          input_distribution=QuantOps.ActHParams.InputDistribution.SYMMETRIC,
          prec=attention_out_inputs_prec,
          half_shift=half_shift,
          bounds=attention_out_inputs_hyper),
      quant_type=quant_type,
      weight_quant_granularity=quant_config.QuantGranularity.PER_CHANNEL)

  # TODO(b/169178846): We are quantizing attention_act_probs with unsigned
  # fixed bounds equal to [0.0, 1.0] to match the range of softmax, since
  # attn_act_probs is Softmax(Q*K) .
  attention_act_probs_inputs_hyper = 1.0
  dot_product_attention = aqt_flax_attention.DotProductAttnHParams(
      attn_act_q=QuantOps.ActHParams(
          QuantOps.ActHParams.InputDistribution.SYMMETRIC,
          prec=attention_act_q_inputs_prec,
          half_shift=half_shift,
          bounds=attention_act_q_inputs_hyper),
      attn_act_k=QuantOps.ActHParams(
          QuantOps.ActHParams.InputDistribution.SYMMETRIC,
          prec=attention_act_k_inputs_prec,
          half_shift=half_shift,
          bounds=attention_act_k_inputs_hyper),
      attn_act_probs=QuantOps.ActHParams(
          QuantOps.ActHParams.InputDistribution.POSITIVE,
          prec=attention_act_probs_inputs_prec,
          half_shift=half_shift,
          bounds=attention_act_probs_inputs_hyper),
      attn_act_v=QuantOps.ActHParams(
          QuantOps.ActHParams.InputDistribution.SYMMETRIC,
          prec=attention_act_v_inputs_prec,
          half_shift=half_shift,
          bounds=attention_act_v_inputs_hyper),
      quant_type=quant_type,
      softmax=None)
  # TODO(malmaud): Move these other create_from* methods to this file.

  attention = aqt_flax_attention.MultiHeadDotProductAttentionAqt.HParams(
      dense_kqv=attention_kqv_dense,
      dense_out=attention_out_dense,
      attn_acts=dot_product_attention)
  dense_1 = aqt_flax_layers.DenseAqt.HParams(
      weight_prec=mlp_weight_prec,
      weight_half_shift=half_shift,
      quant_act=QuantOps.ActHParams(
          input_distribution=QuantOps.ActHParams.InputDistribution.SYMMETRIC,
          prec=mlp_signed_inputs_prec,
          half_shift=half_shift,
          bounds=mlp_signed_inputs_hyper),
      quant_type=quant_type,
      weight_quant_granularity=quant_config.QuantGranularity.PER_CHANNEL)
  dense_2 = aqt_flax_layers.DenseAqt.HParams(
      weight_prec=mlp_weight_prec,
      weight_half_shift=half_shift,
      quant_act=QuantOps.ActHParams(
          input_distribution=QuantOps.ActHParams.InputDistribution.POSITIVE,
          prec=mlp_pos_inputs_prec,
          half_shift=half_shift,
          bounds=mlp_pos_inputs_hyper),
      quant_type=quant_type,
      weight_quant_granularity=quant_config.QuantGranularity.PER_CHANNEL)
  layer_norm = aqt_flax_layers.LayerNormAqt.HParams(quant_hparams=None)
  mlp_block = models.MlpBlock.HParams(dense_1=dense_1, dense_2=dense_2)
  encoder_block = models.Encoder1DBlock.HParams(
      mlp_block, attention, layer_norm=layer_norm)
  decoder_block = models.EncoderDecoder1DBlock.HParams(
      mlp_block,
      self_attention=attention,
      enc_dec_attention=attention,
      layer_norm=layer_norm)
  encoder = models.Encoder.HParams.create_from_block_template(
      embedding=embedding,
      block_template=encoder_block,
      num_layers=num_layers,
      layer_norm=layer_norm)
  if logits_via_embeddings:
    logits_hparams = None
  else:
    logits_hparams = aqt_flax_layers.DenseAqt.HParams(
        weight_prec=embedding_weight_prec,
        weight_half_shift=half_shift,
        quant_act=QuantOps.ActHParams(
            input_distribution=QuantOps.ActHParams.InputDistribution.SYMMETRIC,
            prec=logits_inputs_prec,
            half_shift=half_shift,
            bounds=logits_inputs_hyper),
        quant_type=quant_type,
        weight_quant_granularity=quant_config.QuantGranularity.PER_CHANNEL)
  decoder = models.Decoder.HParams.create_from_block_template(
      embedding=embedding,
      block_template=decoder_block,
      num_layers=num_layers,
      logits=logits_hparams,
      layer_norm=layer_norm)
  return models.Transformer.HParams(
      encoder=encoder,
      decoder=decoder,
      emb_dim=emb_dim,
      num_heads=num_heads,
      qkv_dim=qkv_dim,
      mlp_dim=mlp_dim,
      share_embeddings=True,
      logits_via_embedding=logits_via_embeddings,
  )


def create_training_hparams_from_base_config(
    base_config: BaseConfig) -> training_hparams.TrainingHParams:
  """Creates TrainingHParams instance from BaseConfig instance."""

  num_layers, emb_dim, num_heads, qkv_dim, mlp_dim = None, None, None, None, None
  if base_config.size == BaseConfigSize.MINIMAL_MODEL:
    num_layers = 1
    emb_dim = 1
    num_heads = 1
    qkv_dim = 1
    mlp_dim = 1
  elif base_config.size == BaseConfigSize.SMALL_MODEL:
    num_layers = 3
    emb_dim = 512
    num_heads = 8
    qkv_dim = 512
    mlp_dim = 2048
  elif base_config.size == BaseConfigSize.FULL_MODEL:
    num_layers = 6
    emb_dim = 1024
    num_heads = 16
    qkv_dim = 1024
    mlp_dim = 4096
  else:
    raise ValueError(f'Unknown base config size {base_config.size}')

  prec = base_config.prec

  # Check that precision and quantization target are consistent
  if prec is None and base_config.quant_target != BaseConfigQuantTarget.NONE:
    raise ValueError(
        f'Cannot set quantization target to {base_config.quant_target} when requested precision is None.'
    )

  if base_config.quant_target == BaseConfigQuantTarget.NONE and prec is not None:
    raise ValueError(
        f'Cannot set quantization precision to {prec} when quantization target is NONE.'
    )

  # This variable keeps track of whether we know how to handle the
  # quantization target. If we don't, we want to raise an error to avoid
  # silently using defaults.
  recognized_quant_target = False
  if base_config.quant_target == BaseConfigQuantTarget.NONE:
    recognized_quant_target = True

  # Weight quantization
  mlp_weight_prec = None
  embedding_weight_prec = None
  attention_weight_prec = None
  if base_config.quant_target in [
      BaseConfigQuantTarget.WEIGHTS_ONLY,
      BaseConfigQuantTarget.WEIGHTS_AND_SOME_ACTS_1,
      BaseConfigQuantTarget.WEIGHTS_AND_SOME_ACTS_2,
      BaseConfigQuantTarget.WEIGHTS_AND_ACTS,
  ]:
    recognized_quant_target = True
    mlp_weight_prec = prec
    embedding_weight_prec = prec
    attention_weight_prec = prec

  # Activation quantization
  mlp_pos_inputs_prec = None
  mlp_pos_inputs_hyper_fixed_bound = None
  mlp_signed_inputs_prec = None
  mlp_signed_inputs_hyper_fixed_bound = None
  attention_kqv_inputs_prec = None
  attention_kqv_inputs_hyper_fixed_bound = None
  attention_out_inputs_prec = None
  attention_out_inputs_hyper_fixed_bound = None
  logits_inputs_prec = None
  logits_inputs_hyper_fixed_bound = None
  attention_act_q_inputs_prec = None
  attention_act_q_inputs_hyper_fixed_bound = None
  attention_act_k_inputs_prec = None
  attention_act_k_inputs_hyper_fixed_bound = None
  attention_act_probs_inputs_prec = None
  attention_act_v_inputs_prec = None
  attention_act_v_inputs_hyper_fixed_bound = None
  if base_config.quant_target in [
      BaseConfigQuantTarget.WEIGHTS_AND_SOME_ACTS_1,
      BaseConfigQuantTarget.WEIGHTS_AND_SOME_ACTS_2,
      BaseConfigQuantTarget.WEIGHTS_AND_ACTS,
  ]:
    recognized_quant_target = True
    mlp_pos_inputs_prec = prec
    mlp_signed_inputs_prec = prec
    attention_kqv_inputs_prec = prec
    attention_out_inputs_prec = prec

    mlp_signed_inputs_hyper_fixed_bound = DEFAULTS['mlp_dense_1']['fixed']
    mlp_pos_inputs_hyper_fixed_bound = DEFAULTS['mlp_dense_2']['fixed']
    attention_kqv_inputs_hyper_fixed_bound = DEFAULTS['att_dense_kqv']['fixed']
    attention_out_inputs_hyper_fixed_bound = DEFAULTS['att_dense_out']['fixed']

  if base_config.quant_target in [
      BaseConfigQuantTarget.WEIGHTS_AND_SOME_ACTS_2,
      BaseConfigQuantTarget.WEIGHTS_AND_ACTS,
  ]:
    logits_inputs_prec = prec
    logits_inputs_hyper_fixed_bound = DEFAULTS['logits']['fixed']

  if base_config.quant_target == BaseConfigQuantTarget.WEIGHTS_AND_ACTS:
    # TODO(shivaniagrawal): experimentally find reasonable default bounds.
    attention_act_q_inputs_prec = prec
    attention_act_q_inputs_hyper_fixed_bound = DEFAULTS['attn_act_q']['fixed']
    attention_act_k_inputs_prec = prec
    attention_act_k_inputs_hyper_fixed_bound = DEFAULTS['attn_act_k']['fixed']
    attention_act_probs_inputs_prec = prec
    attention_act_v_inputs_prec = prec
    attention_act_v_inputs_hyper_fixed_bound = DEFAULTS['attn_act_v']['fixed']

  if not recognized_quant_target:
    raise ValueError(f'Unknown quantization target {base_config.quant_target}')

  learning_rate_scheduler_hparams = training_hparams.LearningRateSchedulerHParams(
      factors='constant * linear_warmup * rsqrt_decay',
      base_learning_rate=0.0625,
      warmup_steps=1000,
      decay_factor=0.5,
      steps_per_decay=20000,
      steps_per_cycle=100000)
  batch_size = 256
  num_train_steps = 200000
  weight_decay = 0.0
  beta1 = .9
  beta2 = .98
  eps = 1e-9
  random_seed = 0
  hardware_rng = True

  if base_config.quant_target == BaseConfigQuantTarget.NONE:
    weight_outlier_regularization = 0.0
  else:
    weight_outlier_regularization = 1.0
  weight_outlier_regularization_regex = '^.*kernel$'

  prefer_int8_to_int32_dot = True
  quant_type = QuantType.FAKE_QUANT

  activation_bound_update_freq = -1
  activation_bound_start_step = -1

  model_hparams = create_base_transformer_hparams(
      mlp_weight_prec=mlp_weight_prec,
      embedding_weight_prec=embedding_weight_prec,
      attention_weight_prec=attention_weight_prec,
      num_layers=num_layers,
      num_heads=num_heads,
      emb_dim=emb_dim,
      qkv_dim=qkv_dim,
      mlp_dim=mlp_dim,
      mlp_pos_inputs_prec=mlp_pos_inputs_prec,
      mlp_pos_inputs_hyper=mlp_pos_inputs_hyper_fixed_bound,
      mlp_signed_inputs_prec=mlp_signed_inputs_prec,
      mlp_signed_inputs_hyper=mlp_signed_inputs_hyper_fixed_bound,
      attention_kqv_inputs_prec=attention_kqv_inputs_prec,
      attention_kqv_inputs_hyper=attention_kqv_inputs_hyper_fixed_bound,
      attention_out_inputs_prec=attention_out_inputs_prec,
      attention_out_inputs_hyper=attention_out_inputs_hyper_fixed_bound,
      attention_act_q_inputs_prec=attention_act_q_inputs_prec,
      attention_act_q_inputs_hyper=attention_act_q_inputs_hyper_fixed_bound,
      attention_act_k_inputs_prec=attention_act_k_inputs_prec,
      attention_act_k_inputs_hyper=attention_act_k_inputs_hyper_fixed_bound,
      attention_act_probs_inputs_prec=attention_act_probs_inputs_prec,
      attention_act_v_inputs_prec=attention_act_v_inputs_prec,
      attention_act_v_inputs_hyper=attention_act_v_inputs_hyper_fixed_bound,
      logits_inputs_prec=logits_inputs_prec,
      logits_inputs_hyper=logits_inputs_hyper_fixed_bound,
      logits_via_embeddings=True,
      quant_type=quant_type,
  )

  metadata = os_hparams_utils.HParamsMetadata(
      description='', last_updated_time=time.time())

  return training_hparams.TrainingHParams(
      metadata=metadata,
      model_hparams=model_hparams,
      learning_rate_schedule=learning_rate_scheduler_hparams,
      weight_decay=weight_decay,
      per_host_batch_size=batch_size,
      num_train_steps=num_train_steps,
      beta1=beta1,
      beta2=beta2,
      eps=eps,
      random_seed=random_seed,
      hardware_rng=hardware_rng,
      activation_bound_update_freq=activation_bound_update_freq,
      activation_bound_start_step=activation_bound_start_step,
      weight_outlier_regularization=weight_outlier_regularization,
      weight_outlier_regularization_regex=weight_outlier_regularization_regex,
      prefer_int8_to_int32_dot=prefer_int8_to_int32_dot,
  )


def create_training_hparams_from_flags() -> training_hparams.TrainingHParams:
  """Creates TrainingHParams instance from flags."""
  base_config = BaseConfig.create_from_flags()
  return create_training_hparams_from_base_config(base_config)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Mnist example."""

import copy
import functools
import sys
from typing import Any, Callable
from absl import app
from aqt.jax.v2 import config as aqt_config
from aqt.jax.v2 import tiled_dot_general
from aqt.jax.v2 import utils
from aqt.jax.v2.flax import aqt_flax
from aqt.jax.v2.flax import aqt_flax_calibration
from aqt.jax.v2.flax import delayed_scaling_calibration
from flax import linen as nn
from flax import struct
from flax.metrics import tensorboard
import jax
import jax.numpy as jnp
import numpy as np
import optax
import tensorflow_datasets as tfds

CALIBRATION_STATS = delayed_scaling_calibration.CALIBRATION_STATS
Dataset = dict[str, jnp.ndarray]


class CNN(nn.Module):
  """A simple CNN model."""

  bn_use_stats: bool
  aqt_cfg_dg: None | aqt_config.DotGeneral = None
  aqt_cfg_conv: None | aqt_config.DotGeneralRaw = None
  weights_quant_mode: utils.QuantMode = utils.QuantMode.TRAIN
  activation_quant_mode: utils.QuantMode = utils.QuantMode.TRAIN
  use_legacy_freezer: bool = False

  def get_flax_cls(
      self,
      aqt_cfg: aqt_config.DotGeneral | aqt_config.DotGeneralRaw,
      tiling_cfg: None | tiled_dot_general.Cfg = None):
    if isinstance(aqt_cfg, aqt_config.DotGeneral):
      return functools.partial(
          aqt_flax.AqtDotGeneral,
          aqt_cfg,
          lhs_quant_mode=self.activation_quant_mode,
          rhs_quant_mode=self.weights_quant_mode,
          tiling_cfg=tiling_cfg,
          use_legacy_freezer=self.use_legacy_freezer,
          lhs_freeze_mode=aqt_flax.FreezerMode.CALIBRATION,
          rhs_freeze_mode=aqt_flax.FreezerMode.CALIBRATION_AND_VALUE,
          )

    if aqt_cfg is None:
      return None
    aqt_cfg_conv = aqt_config.default_unquantized_config()
    aqt_cfg_conv.fwd = aqt_cfg
    return functools.partial(
        aqt_flax.AqtConvGeneralDilated,
        aqt_cfg_conv,
        lhs_quant_mode=self.activation_quant_mode,
        rhs_quant_mode=self.weights_quant_mode,
        lhs_freeze_mode=aqt_flax.FreezerMode.CALIBRATION,
        rhs_freeze_mode=aqt_flax.FreezerMode.CALIBRATION_AND_VALUE,
    )

  @nn.compact
  def __call__(self, x):
    tiling_cfg = tiled_dot_general.Cfg(
        lhs=tiled_dot_general.TensorTiling(
            contraction_axes=[
                tiled_dot_general.AxisTiling(
                    axis=1, tile_count=2, tile_size=None
                ),
            ],
            remaining_axes=[],
        ),
        rhs=tiled_dot_general.TensorTiling(
            contraction_axes=[
                tiled_dot_general.AxisTiling(
                    axis=0, tile_count=2, tile_size=None
                ),
            ],
            remaining_axes=[],
        ),
    )

    aqt_dg = self.get_flax_cls(self.aqt_cfg_dg, tiling_cfg=tiling_cfg)
    aqt_conv = self.get_flax_cls(self.aqt_cfg_conv)

    use_running_avg = not self.bn_use_stats
    x = nn.Conv(
        features=32, kernel_size=(3, 3), conv_general_dilated_cls=aqt_conv)(x)
    x = nn.BatchNorm(use_running_average=use_running_avg, dtype=x.dtype)(x)
    x = nn.relu(x)
    x = nn.avg_pool(x, window_shape=(2, 2), strides=(2, 2))
    x = nn.Conv(
        features=64, kernel_size=(3, 3), conv_general_dilated_cls=aqt_conv)(x)
    x = nn.BatchNorm(use_running_average=use_running_avg, dtype=x.dtype)(x)
    x = nn.relu(x)
    x = nn.avg_pool(x, window_shape=(2, 2), strides=(2, 2))
    x = x.reshape((x.shape[0], -1))  # flatten
    x = nn.Dense(features=256, dot_general_cls=aqt_dg)(x)
    x = nn.relu(x)
    x = nn.Dense(features=10, dot_general_cls=aqt_dg)(x)

    # Simple demonstration of how to quantize einsum.
    # Since rhs is activation, we swap the configuration for lhs and rhs, since
    # activation's channelwise config should be set to 'per_tensor'.
    aqt_cfg_rhs_activation = copy.deepcopy(self.aqt_cfg_dg)
    if self.aqt_cfg_dg is not None:
      aqt_cfg_rhs_activation.fwd.dg_quantizer.swap_lhs_and_rhs()

    identity = jnp.identity(10, dtype=x.dtype)
    einsum = aqt_flax.AqtEinsum(
        aqt_cfg_rhs_activation,
        lhs_quant_mode=self.weights_quant_mode,
        rhs_quant_mode=self.activation_quant_mode,
        # These assertions are useful when AqtEinsum definition is far away
        # from usage spot (through injection).
        # This is especially useful when specifying tiling.
        assert_eqn='bc,ab->ac',
        assert_lhs_shape=(10, 10),
        assert_rhs_shape=(None, 10),
        tile_sizes={'b': 5},
        use_legacy_freezer=self.use_legacy_freezer,
        lhs_freeze_mode=aqt_flax.FreezerMode.CALIBRATION_AND_VALUE,
        rhs_freeze_mode=aqt_flax.FreezerMode.CALIBRATION,
    )
    # Note for AQT developers:
    #   This equation is harder because jnp.einsum and einsum swap lhs and rhs.
    x = einsum('bc,ab->ac', identity, x)
    return x


@functools.partial(jax.jit, static_argnums=(3,))
def apply_model(model_params, images, labels, apply_fn):
  """Computes gradients, loss and accuracy for a single batch."""

  def loss_fn(model):
    logits, updated_var = apply_fn(
        model, images, rngs={'params': jax.random.PRNGKey(0)}, mutable=True
    )
    one_hot = jax.nn.one_hot(labels, 10)
    loss = jnp.mean(optax.softmax_cross_entropy(logits=logits, labels=one_hot))
    return loss, (logits, updated_var)

  grad_fn = jax.value_and_grad(loss_fn, has_aux=True, allow_int=True)
  aux, grads = grad_fn(model_params)
  loss, (logits, updated_var) = aux
  accuracy = jnp.mean(jnp.argmax(logits, -1) == labels) * 100
  return grads, loss, accuracy, updated_var


def update_model(state, grads, updated_var):
  # CALIBRATION_STATS is a MutableArray collection and updated in place,
  # so it doesn't go through a gradient update (and can't be returned from
  # a jitted function), so we just copy it over to the new state
  calibration_stats = state.model.get(CALIBRATION_STATS)
  state = update_model_params_with_grads(state, grads, updated_var)
  if calibration_stats:
    state = state.replace(
        model=state.model | {CALIBRATION_STATS: calibration_stats}
    )
  return state


@jax.jit
def update_model_params_with_grads(state, grads, updated_var):
  params = state.model['params']
  param_grad = grads['params']
  updates, new_opt_state = state.tx.update(param_grad, state.opt_state, params)
  new_params = optax.apply_updates(params, updates)
  updated_var.update(params=new_params)
  return state.replace(
      model=updated_var,
      opt_state=new_opt_state,
  )


def _prepare_data_perm(ds, batch_size, rng, num_steps=sys.maxsize):
  ds_size = len(ds['image'])
  num_steps = min(num_steps, ds_size // batch_size)

  perms = jax.random.permutation(rng, len(ds['image']))
  perms = perms[: num_steps * batch_size]  # skip incomplete batch
  perms = perms.reshape((num_steps, batch_size))

  return perms


def train_epoch(state, train_ds, batch_size, rng):
  """Train for a single epoch."""
  perms = _prepare_data_perm(train_ds, batch_size, rng)

  epoch_loss = []
  epoch_accuracy = []

  for perm in perms:
    batch_images = train_ds['image'][perm, ...]
    batch_labels = train_ds['label'][perm, ...]
    grads, loss, accuracy, updated_var = apply_model(
        state.model,
        batch_images,
        batch_labels,
        state.cnn_train.apply,
    )
    state = update_model(state, grads, updated_var)
    epoch_loss.append(loss)
    epoch_accuracy.append(accuracy)
  train_loss = np.mean(epoch_loss)
  train_accuracy = np.mean(epoch_accuracy)
  return state, train_loss, train_accuracy


def get_datasets():
  """Load MNIST train and test datasets into memory."""
  print('get_datasets started')
  ds_builder = tfds.builder('mnist')
  ds_builder.download_and_prepare()
  train_ds = tfds.as_numpy(ds_builder.as_dataset(split='train', batch_size=-1))
  test_ds = tfds.as_numpy(ds_builder.as_dataset(split='test', batch_size=-1))
  train_ds['image'] = jnp.float32(train_ds['image']) / 255.0
  test_ds['image'] = jnp.float32(test_ds['image']) / 255.0
  print('get_datasets DONE')
  return train_ds, test_ds


class TrainState(struct.PyTreeNode):
  """Train state."""

  cnn_train: Any = struct.field(pytree_node=False)
  cnn_eval: Any = struct.field(pytree_node=False)
  model: Any = struct.field(pytree_node=True)
  tx: optax.GradientTransformation = struct.field(pytree_node=False)
  opt_state: optax.OptState = struct.field(pytree_node=True)


def create_train_state(rng, aqt_cfg_dg=None, aqt_cfg_conv=None):
  """Creates initial `TrainState`."""
  cnn_train = CNN(
      bn_use_stats=True, aqt_cfg_dg=aqt_cfg_dg, aqt_cfg_conv=None)
  model = cnn_train.init({'params': rng}, jnp.ones([1, 28, 28, 1]))
  learning_rate = 0.1
  momentum = 0.9
  tx = optax.sgd(learning_rate, momentum)
  cnn_eval = CNN(
      bn_use_stats=False,
      aqt_cfg_dg=aqt_cfg_dg,
      aqt_cfg_conv=aqt_cfg_conv
      )
  return TrainState(
      cnn_train=cnn_train,
      cnn_eval=cnn_eval,
      model=model,
      tx=tx,
      opt_state=tx.init(model['params']),
  )


def train_and_evaluate(
    num_epochs: int,
    workdir: str,
    aqt_cfg_dg: None | aqt_config.DotGeneral = None,
    aqt_cfg_conv: None | aqt_config.DotGeneralRaw = None,
    state: None | TrainState = None,
    datasets: None | tuple[Dataset, Dataset] = None,
) -> TrainState:
  """Execute model training and evaluation loop."""
  if datasets is None:
    train_ds, test_ds = get_datasets()
  else:
    train_ds, test_ds = datasets
  rng = jax.random.key(0)

  summary_writer = tensorboard.SummaryWriter(workdir)

  rng, init_rng = jax.random.split(rng)
  if state is None:
    assert aqt_cfg_dg is not None
    state = create_train_state(init_rng,
                               aqt_cfg_dg=aqt_cfg_dg,
                               aqt_cfg_conv=aqt_cfg_conv
                               )

  batch_size = 128
  for epoch in range(1, num_epochs + 1):
    rng, input_rng = jax.random.split(rng)
    state, train_loss, train_accuracy = train_epoch(
        state, train_ds, batch_size, input_rng
    )
    _, test_loss, test_accuracy, _ = apply_model(
        state.model, test_ds['image'], test_ds['label'], state.cnn_eval.apply
    )

    stats = [
        f'epoch: {epoch:3d}',
        f'{train_loss = :.30f}',
        f'{test_loss  = :.30f}',
        f'{train_accuracy = :34.30f}',
        f'{test_accuracy  = :34.30f}',
    ]
    print('\n'.join(stats) + '\n', flush=True)

    summary_writer.scalar('train_loss', train_loss, epoch)
    summary_writer.scalar('train_accuracy', train_accuracy, epoch)
    summary_writer.scalar('test_loss', test_loss, epoch)
    summary_writer.scalar('test_accuracy', test_accuracy, epoch)

  summary_writer.flush()
  return state


def serving_conversion(
    train_state: TrainState,
    weight_only: bool = True,
    legacy_for_freeze: bool = False,
    legacy_for_serve: bool = False,
) -> tuple[Callable[..., Any], dict[str, Any]]:
  """Model conversion (quantized weights freezing).

  Convert the model parameter for serving. During conversion, quantized weights
  are created as variables, along with their scales.

  Args:
    train_state: TrainState containing model definitions and params.
    weight_only: If set, does not quantize activations.
    legacy_for_freeze: If set, use legacy freezer during freeze.
    legacy_for_serve: If set, use legacy freezer during serve.

  Returns:
    A tuple of serving function, and converted model parameters.
  """
  aqt_cfg_dg = train_state.cnn_eval.aqt_cfg_dg
  aqt_cfg_conv = train_state.cnn_eval.aqt_cfg_conv
  input_shape = (1, 28, 28, 1)
  activation_quant_mode = (
      utils.QuantMode.TRAIN if weight_only else utils.QuantMode.CONVERT
  )
  cnn_freeze = CNN(
      bn_use_stats=False,
      aqt_cfg_dg=aqt_cfg_dg,
      aqt_cfg_conv=aqt_cfg_conv,
      weights_quant_mode=utils.QuantMode.CONVERT,
      activation_quant_mode=activation_quant_mode,
      use_legacy_freezer=legacy_for_freeze,
  )
  _, model_serving = cnn_freeze.apply(
      train_state.model,
      jnp.ones(input_shape),
      rngs={'params': jax.random.PRNGKey(0)},
      mutable=True,
  )

  activation_quant_mode = (
      utils.QuantMode.TRAIN if weight_only else utils.QuantMode.SERVE
  )
  cnn_serve = CNN(
      bn_use_stats=False,
      aqt_cfg_dg=aqt_cfg_dg,
      aqt_cfg_conv=aqt_cfg_conv,
      weights_quant_mode=utils.QuantMode.SERVE,
      activation_quant_mode=activation_quant_mode,
      use_legacy_freezer=legacy_for_serve,
  )

  return cnn_serve.apply, model_serving


def _merge_pytrees(from_model, to_model):
  """Copies the parameters from from_model to to_model."""
  from_model_flattened, _ = jax.tree_util.tree_flatten_with_path(from_model)
  to_model_flattened, to_model_treedef = jax.tree_util.tree_flatten_with_path(
      to_model
  )
  from_model_kp_to_val = {kp: val for kp, val in from_model_flattened}
  merged_flattened = []
  for kp, val in to_model_flattened:
    if kp in from_model_kp_to_val:
      merged_flattened.append((kp, from_model_kp_to_val[kp]))
    else:
      merged_flattened.append((kp, val))

  merged_model = jax.tree_util.tree_unflatten(
      to_model_treedef, [v for _, v in merged_flattened]
  )

  return merged_model


def update_cfg_with_calibration(aqt_cfg):
  """Updates aqt_cfg for static range calibration."""
  sr_calibration_cls = functools.partial(
      aqt_flax_calibration.MeanOfAbsMaxCalibration,
      quant_collection='qc',
  )

  aqt_config.set_fwd_calibration(aqt_cfg, sr_calibration_cls)

  # For static range calibration, the calibration axis for activation should
  # be set to per_tensor, since its dimensions could be different during
  # training and during inference.
  aqt_cfg.fwd.dg_quantizer.lhs.calib_shared_axes = 'per_tensor'


def update_cfg_raw_with_calibration(aqt_cfg_raw):
  """Updates aqt_cfg for static range calibration."""
  sr_calibration_cls = functools.partial(
      aqt_flax_calibration.MeanOfAbsMaxCalibration,
      quant_collection='qc',
  )

  aqt_cfg_raw.dg_quantizer.lhs.calibration = sr_calibration_cls
  aqt_cfg_raw.dg_quantizer.rhs.calibration = sr_calibration_cls

  aqt_cfg_raw.dg_quantizer.lhs.calib_shared_axes = 'per_tensor'


def calibration_conversion(
    train_state: TrainState,
) -> tuple[Callable[..., Any], dict[str, Any]]:
  """Model conversion (initializing calibration parameters).

  Newly initialize variables to store the quantization statistics collected
  during calibration process.

  Args:
    train_state: TrainState containing model definitions and params.

  Returns:
    A tuple of calibration function, and an updated model parameters with new
    variables to store the gathered quantization statistics.
  """
  cnn_calibrate = CNN(
      bn_use_stats=False,
      aqt_cfg_dg=train_state.cnn_eval.aqt_cfg_dg,
      aqt_cfg_conv=train_state.cnn_eval.aqt_cfg_conv,
      # Both side should be calibrated.
      weights_quant_mode=utils.QuantMode.CALIBRATE,
      activation_quant_mode=utils.QuantMode.CALIBRATE,
  )

  # Initialize the model, and then load the checkpoint into the initialized
  # parameter dict.
  model_calibrated_init = cnn_calibrate.init(
      jax.random.PRNGKey(0), jnp.ones([1, 28, 28, 1])
  )
  model_calibrated = _merge_pytrees(train_state.model, model_calibrated_init)

  return cnn_calibrate.apply, model_calibrated


def calibrate_epoch(
    calibrate_func,
    model_calibrated,
    train_ds,
    batch_size,
    rng,
    calibration_steps,
):
  """Calibrates for a single epoch."""
  perms = _prepare_data_perm(train_ds, batch_size, rng, calibration_steps)

  for perm in perms:
    batch_images = train_ds['image'][perm, ...]
    batch_labels = train_ds['label'][perm, ...]

    # Calibration simply updates model during inference; it does NOT apply any
    # gradients.
    _, _, _, model_calibrated = apply_model(
        model_calibrated, batch_images, batch_labels, calibrate_func
    )

  return model_calibrated


def calibrate(state: TrainState, calibration_steps: int) -> TrainState:
  """Calibrate."""
  train_ds, _ = get_datasets()
  rng = jax.random.key(0)
  batch_size = 128
  calibration_func, model_calibrated = calibration_conversion(state)

  model_calibrated = calibrate_epoch(
      calibration_func,
      model_calibrated,
      train_ds,
      batch_size,
      rng,
      calibration_steps,
  )

  return state.replace(model=model_calibrated)


@functools.partial(jax.jit, static_argnums=(1,))
def serve(state: TrainState, weight_only: bool = True):
  """Take train state, freeze integer weights, and serve."""
  # get sample serving data
  _, test_ds = get_datasets()
  sample_image, sample_label = test_ds['image'][:64], test_ds['label'][:64]
  # serving
  serve_fn, model_serving = serving_conversion(state, weight_only=weight_only)
  logits = serve_fn(
      model_serving, sample_image, rngs={'params': jax.random.PRNGKey(0)}
  )
  # compute serving loss
  one_hot = jax.nn.one_hot(sample_label, 10)
  loss = jnp.mean(optax.softmax_cross_entropy(logits=logits, labels=one_hot))
  return loss


def serve_fn_hlo(state):
  """Example on how to inspect HLO to verify if everything was quantized."""
  # get sample serving data
  _, test_ds = get_datasets()
  sample_image = test_ds['image'][:64]
  # serving
  serve_fn, model_serving = serving_conversion(state)
  # The following XLA graph is only needed for debugging purpose
  hlo = (
      jax.jit(serve_fn)
      .lower(
          model_serving,
          sample_image,
          rngs={'params': jax.random.PRNGKey(0)},
      )
      .compiler_ir('hlo')
      .as_hlo_module()
  )
  return hlo


def main(argv):
  del argv

  # 1. TRAIN.
  aqt_cfg_dg = aqt_config.fully_quantized(fwd_bits=8, bwd_bits=8)
  state = train_and_evaluate(
      num_epochs=1, workdir='/tmp/aqt_mnist_example', aqt_cfg_dg=aqt_cfg_dg
  )

  # 2. Calibration.
  update_cfg_with_calibration(state.cnn_train.aqt_cfg_dg)
  update_cfg_with_calibration(state.cnn_eval.aqt_cfg_dg)
  state = calibrate(state, calibration_steps=10)

  # 3. TRAIN with the calibrated stats.
  state = train_and_evaluate(
      num_epochs=1, workdir='/tmp/aqt_mnist_example', state=state
  )

  # 4. CONVERT & SERVE.
  loss = serve(state, weight_only=False)
  print('serve loss on sample ds: {}'.format(loss))


if __name__ == '__main__':
  app.run(main)
